#!/usr/bin/env python3
"""SC-Controller - Action Editor - Gyro -> Per Axis component
"""

import itertools
import logging

from scc.actions import Action, AxisAction, GyroAbsAction, GyroAction, MultiAction, NoAction, RangeOP
from scc.constants import STICK, SCButtons
from scc.gui.ae import AEComponent, describe_action
from scc.gui.ae.gyro_action import TRIGGERS, fill_buttons, is_gyro_enable
from scc.gui.simple_chooser import SimpleChooser
from scc.modifiers import ModeModifier
from scc.tools import _, ensure_size, nameof

log = logging.getLogger("AE.Gyro")

__all__ = ["GyroComponent"]


class GyroComponent(AEComponent):
	GLADE = "ae/gyro.glade"
	NAME = "gyro"
	CTXS = Action.AC_GYRO
	PRIORITY = 2

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)
		self._recursing = False
		self.axes = [None, None, None]

	def load(self):
		if self.loaded:
			return
		AEComponent.load(self)
		cbGyroButton = self.builder.get_object("cbGyroButton")
		self._recursing = True
		cbGyroButton = self.builder.get_object("cbGyroButton")
		fill_buttons(cbGyroButton)
		self._recursing = False
		self.buttons = [self.builder.get_object(x) for x in ("btPitch", "btYaw", "btRoll")]
		self.cbs = [self.builder.get_object(x) for x in ("cbPitchAbs", "cbYawAbs", "cbRollAbs")]
		self.labels = [self.builder.get_object(x) for x in ("lblPitch", "lblYaw", "lblRoll")]

	def set_action(self, mode, action):
		if self.handles(mode, action):
			if isinstance(action, ModeModifier):
				self._recursing = True
				self.builder.get_object("cbInvertGyro").set_active(bool(action.default))
				self._recursing = False
				b = next(itertools.islice(action.mods.keys(), 0, 1))
				action = action.mods[b] or action.default
				self.select_gyro_button(b)
			else:
				self.select_gyro_button(None)

			actions = [action]
			if isinstance(action, MultiAction):
				actions = action.actions

			self._recursing = True
			for a in actions:
				if isinstance(a, GyroAction):
					pars = ensure_size(3, a.parameters)
					for i in range(3):
						if pars[i] is not None:
							self.axes[i] = pars[i]
							self.cbs[i].set_active(isinstance(a, GyroAbsAction))
			self.update()
			self._recursing = False

	def get_button_title(self):
		return _("Per Axis")

	def handles(self, mode, action):
		if is_gyro_enable(action):
			action = next(itertools.islice(action.mods.values(), 0, 1))
		if isinstance(action, GyroAction):  # Takes GyroAbsAction as well
			return True
		if isinstance(action, MultiAction):
			for a in action.actions:
				if not isinstance(a, GyroAction):
					return False
			return True
		return False

	def on_select_axis(self, source, *a):
		i = self.buttons.index(source)

		def cb(action):
			self.axes[i] = action.parameters[0]
			self.update()
			self.send()

		b = SimpleChooser(self.app, "axis", cb)
		b.set_title(_("Select Axis"))
		b.hide_mouse()
		b.display_action(Action.AC_STICK, AxisAction(self.axes[i]))
		b.show(self.editor.window)

	def on_abs_changed(self, source, *a):
		if self._recursing:
			return
		self.send()

	def select_gyro_button(self, item):
		"""Just sets combobox value"""
		cb = self.builder.get_object("cbGyroButton")
		rvSoftLevel = self.builder.get_object("rvSoftLevel")
		sclSoftLevel = self.builder.get_object("sclSoftLevel")
		lblSoftLevel = self.builder.get_object("lblSoftLevel")
		model = cb.get_model()
		self._recursing = True
		button = None
		if isinstance(item, RangeOP):
			button = nameof(item.what)
			sclSoftLevel.set_value(item.value)
			rvSoftLevel.set_reveal_child(True)
			if item.what == STICK:
				lblSoftLevel.set_label(_("Stick deadzone"))
			else:
				lblSoftLevel.set_label(_("Trigger Pull Level"))
		elif item is not None:
			button = nameof(item.name)
		for row in model:
			if button == row[0] and row[1] != None:
				cb.set_active_iter(row.iter)
				self._recursing = False
				return
		self._recursing = False

	def on_cbInvertGyro_toggled(self, cb, *a):
		lblGyroEnable = self.builder.get_object("lblGyroEnable")
		if cb.get_active():
			lblGyroEnable.set_label(_("Gyro Disable Button"))
		else:
			lblGyroEnable.set_label(_("Gyro Enable Button"))
		if not self._recursing:
			self.send()

	def on_sclSoftLevel_format_value(self, scale, value):
		return "%s%%" % (int(value * 100.0),)

	def update(self, *a):
		for i in range(3):
			self.labels[i].set_label(describe_action(Action.AC_STICK, AxisAction, self.axes[i]))

	def send(self, *a):
		if self._recursing:
			return

		rvSoftLevel = self.builder.get_object("rvSoftLevel")
		sclSoftLevel = self.builder.get_object("sclSoftLevel")
		cbGyroButton = self.builder.get_object("cbGyroButton")
		cbInvertGyro = self.builder.get_object("cbInvertGyro")
		item = cbGyroButton.get_model().get_value(cbGyroButton.get_active_iter(), 0)
		rvSoftLevel.set_reveal_child(item in TRIGGERS)

		normal, n_set = [None, None, None], False
		absolute, a_set = [None, None, None], False

		for i in range(3):
			# Fix case when axis id is zero (ABS_X)
			if self.axes[i] != None:
				if self.cbs[i].get_active():
					absolute[i] = self.axes[i]
					a_set = True
				else:
					normal[i] = self.axes[i]
					n_set = True

		if n_set and a_set:
			action = MultiAction(GyroAction(*normal), GyroAbsAction(*absolute))
		elif n_set:
			action = GyroAction(*normal)
		elif a_set:
			action = GyroAbsAction(*absolute)
		else:
			action = NoAction()

		if item and action:
			what = getattr(SCButtons, item)
			if item in TRIGGERS:
				what = RangeOP(what, ">=", sclSoftLevel.get_value())
			if cbInvertGyro.get_active():
				action = ModeModifier(what, NoAction(), action)
			else:
				action = ModeModifier(what, action)

		self.editor.set_action(action)



#!/usr/bin/env python3
"""SC-Controller - OSD Dialog

Display dialog with text and set of items that user can navigate through and
prints chosen item id to stdout
"""

import logging
import sys

from gi.repository import GdkX11, Gtk

from scc.config import Config
from scc.constants import DEFAULT
from scc.gui.daemon_manager import DaemonManager
from scc.lib import xwrappers as X
from scc.menu_data import MenuData
from scc.osd import OSDWindow, StickController

log = logging.getLogger("osd.dialog")


class Dialog(OSDWindow):
	EPILOG = """Exit codes:
   0  - clean exit, user selected option
  -1  - clean exit, user canceled dialog
   1  - error, invalid arguments
   2  - error, failed to access sc-daemon, sc-daemon reported error or died while dialog is displayed.
   3  - erorr, failed to lock input stick, pad or button(s)
	"""

	def __init__(self, cls="osd-menu"):
		self._buttons = None
		self._text = None

		OSDWindow.__init__(self, cls)
		self.daemon = None
		self.config = None
		self.feedback = None
		self.controller = None
		self.xdisplay = X.Display(hash(GdkX11.x11_get_default_xdisplay()))  # Magic

		self.parent = self.create_parent()
		self.f = Gtk.Fixed()
		self.f.add(self.parent)
		self.add(self.f)

		self._scon = StickController()
		self._scon.connect("direction", self.on_stick_direction)
		self._selected = None
		self._eh_ids = []

	def create_parent(self):
		self._text = Gtk.Label()
		self._buttons = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
		dialog = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
		dialog.pack_start(self._text, True, True, 0)
		dialog.pack_start(self._buttons, True, True, 0)

		dialog.set_name("osd-dialog")
		self._buttons.set_name("osd-dialog-buttons")
		self._text.set_name("osd-dialog-text")
		return dialog

	def pack_items(self, parent, items):
		for item in items:
			if hasattr(item.widget, "set_alignment"):
				item.widget.set_alignment(0.5, 0.5)
			self._buttons.pack_end(item.widget, True, True, 0)

	def use_daemon(self, d):
		"""Allows (re)using already existing DaemonManager instance in same process.
		use_config() should be be called before parse_argumets() if this is used.
		"""
		self.daemon = d
		self._connect_handlers()
		self.on_daemon_connected(self.daemon)

	def use_config(self, c):
		"""Allows reusing already existin Config instance in same process.
		Has to be called before parse_argumets()
		"""
		self.config = c

	def get_menuid(self):
		# Just to be compatibile with menus when called from scc-osd-daemon
		return None

	def get_selected_item_id(self):
		"""Returns ID of selected item or None if nothing is selected.
		"""
		if self._selected:
			return self._selected.id
		return None

	def _add_arguments(self):
		OSDWindow._add_arguments(self)
		self.argparser.add_argument(
			"--confirm-with", type=str, metavar="button", default=DEFAULT, help="button used to confirm choice",
		)
		self.argparser.add_argument(
			"--cancel-with", type=str, metavar="button", default=DEFAULT, help="button used to cancel dialog",
		)
		self.argparser.add_argument(
			"--feedback-amplitude",
			type=int,
			help="enables and sets power of feedback effect generated when active menu option is changed",
		)
		self.argparser.add_argument("--text", type=str, metavar="text", help="Dialog text")
		self.argparser.add_argument("items", type=str, nargs="*", metavar="id text", help="Dialog buttons")

	def parse_argumets(self, argv):
		if not OSDWindow.parse_argumets(self, argv):
			return False
		if not self.config:
			self.config = Config()

		try:
			self.items = MenuData.from_args(self.args.items)
			self._menuid = None
		except ValueError:
			print("%s: error: invalid number of arguments" % (sys.argv[0]), file=sys.stderr)
			return False

		self._text.set_label(self.args.text)

		if self.args.feedback_amplitude:
			side = "LEFT"
			self.feedback = side, int(self.args.feedback_amplitude)

		# Create buttons that are displayed on screen
		items = self.items.generate(self)
		self.items = []
		for item in items:
			item.widget = self.generate_widget(item)
			if item.widget is not None:
				self.items.append(item)
		self.pack_items(self.parent, self.items)
		if len(self.items) == 0:
			print("%s: error: no items in menu" % (sys.argv[0]), file=sys.stderr)
			return False

		return True

	def generate_widget(self, item):
		"""Generates gtk widget for specified menutitem"""
		widget = Gtk.Button.new_with_label(item.label)
		widget.set_relief(Gtk.ReliefStyle.NONE)
		if hasattr(widget.get_children()[0], "set_xalign"):
			widget.get_children()[0].set_xalign(0)
		else:
			widget.get_children()[0].set_halign(Gtk.Align.START)
		widget.set_name("osd-menu-item")

		return widget

	def select(self, index):
		if self._selected:
			self._selected.widget.set_name(self._selected.widget.get_name().replace("-selected", ""))
		if self.items[index].id:
			if self._selected != self.items[index]:
				if self.feedback and self.controller:
					self.controller.feedback(*self.feedback)
			self._selected = self.items[index]
			self._selected.widget.set_name(self._selected.widget.get_name() + "-selected")
			return True
		return False

	def _connect_handlers(self):
		self._eh_ids += [
			(self.daemon, self.daemon.connect("dead", self.on_daemon_died)),
			(self.daemon, self.daemon.connect("error", self.on_daemon_died)),
			(self.daemon, self.daemon.connect("alive", self.on_daemon_connected)),
		]

	def run(self):
		self.daemon = DaemonManager()
		self._connect_handlers()
		OSDWindow.run(self)

	def show(self, *a):
		if not self.select(0):
			self.next_item(1)
		OSDWindow.show(self, *a)

	def on_daemon_died(self, *a):
		log.error("Daemon died")
		self.quit(2)

	def on_failed_to_lock(self, error):
		log.error("Failed to lock input: %s", error)
		self.quit(3)

	def on_daemon_connected(self, *a):
		def success(*a):
			log.error("Sucessfully locked input")

		if not self.config:
			self.config = Config()
		self.controller = self.choose_controller(self.daemon)
		if self.controller is None or not self.controller.is_connected():
			# There is no controller connected to daemon
			self.on_failed_to_lock("Controller not connected")
			return

		ccfg = self.config.get_controller_config(self.controller.get_id())
		self._control_with = ccfg["menu_control"]
		self._confirm_with = ccfg["menu_confirm"] if self.args.confirm_with == DEFAULT else self.args.confirm_with
		self._cancel_with = ccfg["menu_cancel"] if self.args.cancel_with == DEFAULT else self.args.cancel_with

		self._eh_ids += [(self.controller, self.controller.connect("event", self.on_event))]
		locks = [self._control_with, self._confirm_with, self._cancel_with]
		self.controller.lock(success, self.on_failed_to_lock, *locks)

	def quit(self, code=-2):
		if self.get_controller():
			self.get_controller().unlock_all()
		for source, eid in self._eh_ids:
			source.disconnect(eid)
		self._eh_ids = []
		OSDWindow.quit(self, code)

	def next_item(self, direction):
		"""Selects next menu item, based on self._direction"""
		start, i = -1, 0
		try:
			start = self.items.index(self._selected)
			i = start + direction
		except:
			pass
		while True:
			if i == start:
				# Cannot find valid menu item
				self.select(start)
				break
			if i >= len(self.items):
				i = 0
				continue
			if i < 0:
				i = len(self.items) - 1
				continue
			if self.select(i):
				break
			i += direction
			start = max(start, 0)

	def on_stick_direction(self, trash, x, y):
		if x != 0:
			self.next_item(x)

	def on_event(self, daemon, what, data):
		if what == self._control_with:
			self._scon.set_stick(*data)
		elif what == self._cancel_with:
			if data[0] == 0:  # Button released
				self.quit(-1)
		elif what == self._confirm_with:
			if data[0] == 0:  # Button released
				if self._selected and self._selected.callback:
					self._selected.callback(self, self.daemon, self.controller, self._selected)
				elif self._selected:
					self.quit(0)
				else:
					self.quit(-1)



"""hidparse_data - long enums moved from hidparse.py.

Licensed under GPL 2.0
"""

from enum import IntEnum


class MainItem(IntEnum):
	Input = 0x80
	Output = 0x90
	Collection = 0xA0
	Feature = 0xB0
	EndCollection = 0xC0


class GlobalItem(IntEnum):
	UsagePage = 0x00
	LogicalMinimum = 0x10
	LogicalMaximum = 0x20
	PhysicalMinimum = 0x30
	PhysicalMaximum = 0x40
	UnitExponent = 0x50
	Unit = 0x60
	ReportSize = 0x70
	ReportID = 0x80
	ReportCount = 0x90
	Push = 0xA0
	Pop = 0xB0


class LocalItem(IntEnum):
	Usage = 0x00
	UsageMinimum = 0x10
	UsageMaximum = 0x20
	DesignatorIndex = 0x30
	DesignatorMinimum = 0x40
	DesignatorMaximum = 0x50
	StringIndex = 0x70
	StringMinimum = 0x80
	StringMaximum = 0x90
	Delimiter = 0xA0


class Collection(IntEnum):
	Physical = 0x00
	Application = 0x01
	Logical = 0x02
	Report = 0x03
	NamedArray = 0x04
	UsageSwitch = 0x05
	UsageModifier = 0x06


class GenericDesktopPage(IntEnum):
	Undefined = 0x00
	Pointer = 0x01
	Mouse = 0x02
	Reserved = 0x03
	Joystick = 0x04
	GamePad = 0x05
	Keyboard = 0x06
	Keypad = 0x07
	MultiAxisController = 0x08
	TabletPCSystemControls = 0x09
	X = 0x30
	Y = 0x31
	Z = 0x32
	Rx = 0x33
	Ry = 0x34
	Rz = 0x35
	Slider = 0x36
	Dial = 0x37
	Wheel = 0x38
	Hatswitch = 0x39
	CountedBuffer = 0x3A
	ByteCount = 0x3B
	MotionWakeup = 0x3C
	Start = 0x3D
	Select = 0x3E
	Vx = 0x40
	Vy = 0x41
	Vz = 0x42
	Vbrx = 0x43
	Vbry = 0x44
	Vbrz = 0x45
	Vno = 0x46
	FeatureNotification = 0x47
	ResolutionMultiplier = 0x48
	SystemControl = 0x80
	SystemPowerDown = 0x81
	SystemSleep = 0x82
	SystemWakeUp = 0x83
	SystemContextMenu = 0x84
	SystemMainMenu = 0x85
	SystemAppMenu = 0x86
	SystemMenuHelp = 0x87
	SystemMenuExit = 0x88
	SystemMenuSelect = 0x89
	SystemMenuRight = 0x8A
	SystemMenuLeft = 0x8B
	SystemMenuUp = 0x8C
	SystemMenuDown = 0x8D
	SystemColdRestart = 0x8E
	SystemWarmRestart = 0x8F
	DPadUp = 0x90
	DPadDown = 0x91
	DPadRight = 0x92
	DPadLeft = 0x93
	SystemDock = 0xA0
	SystemUndock = 0xA1
	SystemSetup = 0xA2
	SystemBreak = 0xA3
	SystemDebuggerBreak = 0xA4
	ApplicationBreak = 0xA5
	ApplicationDebuggerBreak = 0xA6
	SystemSpeakerMute = 0xA7
	SystemHibernate = 0xA8
	SystemDisplayInvert = 0xB0
	SystemDisplayInternal = 0xB1
	SystemDisplayExternal = 0xB2
	SystemDisplayBoth = 0xB3
	SystemDisplayDual = 0xB4
	SystemDisplayToggleIntExt = 0xB5
	SystemDisplaySwapPrimarySecondary = 0xB6
	SystemDisplayLCDAutoscale = 0xB7


class SimulationControlsPage(IntEnum):
	Undefined = 0x00
	FlightSimulationDevice = 0x01
	AutomobileSimulationDevice = 0x02
	TankSimulationDevice = 0x03
	SpaceshipSimulationDevice = 0x04
	SubmarineSimulationDevice = 0x05
	SailingSimulationDevice = 0x06
	MotorcycleSimulationDevice = 0x07
	SportsSimulationDevice = 0x08
	AirplaneSimulationDevice = 0x09
	HelicopterSimulationDevice = 0x0A
	MagicCarpetSimulationDevice = 0x0B
	BicycleSimulationDevice = 0x0C
	FlightControlStick = 0x20
	FlightStick = 0x21
	CyclicControl = 0x22
	CyclicTrim = 0x23
	FlightYoke = 0x24
	TrackControl = 0x25
	Aileron = 0xB0
	AileronTrim = 0xB1
	AntiTorqueControl = 0xB2
	AutopilotEnable = 0xB3
	ChaffRelease = 0xB4
	CollectiveControl = 0xB5
	DiveBrake = 0xB6
	ElectronicCountermeasures = 0xB7
	Elevator = 0xB8
	ElevatorTrim = 0xB9
	Rudder = 0xBA
	Throttle = 0xBB
	FlightCommunications = 0xBC
	FlareRelease = 0xBD
	LandingGear = 0xBE
	ToeBrake = 0xBF
	Trigger = 0xC0
	WeaponsArm = 0xC1
	WeaponsSelect = 0xC2
	WingFlaps = 0xC3
	Accelerator = 0xC4
	Brake = 0xC5
	Clutch = 0xC6
	Shifter = 0xC7
	Steering = 0xC8
	TurretDirection = 0xC9
	BarrelElevation = 0xCA
	DivePlane = 0xCB
	Ballast = 0xCC
	BicycleCrank = 0xCD
	HandleBars = 0xCE
	FrontBrake = 0xCF
	RearBrake = 0xD0


class VRControlsPage(IntEnum):
	Unidentified = 0x00
	Belt = 0x01
	BodySuit = 0x02
	Flexor = 0x03
	Glove = 0x04
	HeadTracker = 0x05
	HeadMountedDisplay = 0x06
	HandTracker = 0x07
	Oculometer = 0x08
	Vest = 0x09
	AnimatronicDevice = 0x0A
	StereoEnable = 0x20
	DisplayEnable = 0x21


class SportControlsPage(IntEnum):
	Unidentified = 0x00
	BaseballBat = 0x01
	GolfClub = 0x02
	RowingMachine = 0x03
	Treadmill = 0x04
	Oar = 0x30
	Slope = 0x31
	Rate = 0x32
	StickSpeed = 0x33
	StickFaceAngle = 0x34
	StickHeelToe = 0x35
	StickFollowThrough = 0x36
	StickTempo = 0x37
	StickType = 0x38
	StickHeight = 0x39
	Putter = 0x50
	Iron1 = 0x51
	Iron2 = 0x52
	Iron3 = 0x53
	Iron4 = 0x54
	Iron5 = 0x55
	Iron6 = 0x56
	Iron7 = 0x57
	Iron8 = 0x58
	Iron9 = 0x59
	Iron10 = 0x5A
	Iron11 = 0x5B
	SandWedge = 0x5C
	LoftWedge = 0x5D
	PowerWedge = 0x5E
	Wood1 = 0x5F
	Wood3 = 0x60
	Wood5 = 0x61
	Wood7 = 0x62
	Wood9 = 0x63


class GameControlsPage(IntEnum):
	Undefined = 0x00
	GameController3D = 0x01
	PinballDevice = 0x02
	GunDevice = 0x03
	PointofView = 0x20
	TurnRightLeft = 0x21
	PitchForwardBackward = 0x22
	RollRightLeft = 0x23
	MoveRightLeft = 0x24
	MoveForwardBackward = 0x25
	MoveUpDown = 0x26
	LeanRightLeft = 0x27
	LeanForwardBackward = 0x28
	HeightofPOV = 0x29
	Flipper = 0x2A
	SecondaryFlipper = 0x2B
	Bump = 0x2C
	NewGame = 0x2D
	ShootBall = 0x2E
	Player = 0x2F
	GunBolt = 0x30
	GunClip = 0x31
	GunSelector = 0x32
	GunSingleShot = 0x33
	GunBurst = 0x34
	GunAutomatic = 0x35
	GunSafety = 0x36
	GamepadFireJump = 0x37
	GamepadTrigger = 0x39


class GenericDeviceControlsPage(IntEnum):
	Unidentified = 0x00
	BatteryStrength = 0x20
	WirelessChannel = 0x21
	WirelessID = 0x22
	DiscoverWirelessControl = 0x23
	SecurityCodeCharacterEntered = 0x24
	SecurityCodeCharacterErased = 0x25
	SecurityCodeCleared = 0x26


class KeyboardKeypadPage(IntEnum):
	NoEventIndicated = 0x00
	ErrorRollOver = 0x01
	POSTFail = 0x02
	ErrorUndefined = 0x03
	A = 0x04
	B = 0x05
	C = 0x06
	D = 0x07
	E = 0x08
	F = 0x09
	G = 0x0A
	H = 0x0B
	I = 0x0C
	J = 0x0D
	K = 0x0E
	L = 0x0F
	M = 0x10
	N = 0x11
	O = 0x12
	P = 0x13
	Q = 0x14
	R = 0x15
	S = 0x16
	T = 0x17
	U = 0x18
	V = 0x19
	W = 0x1A
	X = 0x1B
	Y = 0x1C
	Z = 0x1D
	Num1 = 0x1E
	Num2 = 0x1F
	Num3 = 0x20
	Num4 = 0x21
	Num5 = 0x22
	Num6 = 0x23
	Num7 = 0x24
	Num8 = 0x25
	Num9 = 0x26
	Num0 = 0x27
	Enter = 0x28
	Escape = 0x29
	Delete = 0x2A
	Tab = 0x2B
	Spacebar = 0x2C
	Minus = 0x2D
	Plus = 0x2E
	LBrace = 0x2F
	RBrace = 0x30
	Backslash = 0x31
	NonUsGrave = 0x32
	Colon = 0x33
	Doublequote = 0x34
	Grave = 0x35
	KeypadComma = 0x36
	Dot = 0x37
	Slash = 0x38
	CapsLock = 0x39
	F1 = 0x3A
	F2 = 0x3B
	F3 = 0x3C
	F4 = 0x3D
	F5 = 0x3E
	F6 = 0x3F
	F7 = 0x40
	F8 = 0x41
	F9 = 0x42
	F10 = 0x43
	F11 = 0x44
	F12 = 0x45
	PrintScreen = 0x46
	ScrollLock = 0x47
	Pause = 0x48
	Insert = 0x49
	Home = 0x4A
	PageUp = 0x4B
	DeleteForward = 0x4C
	End = 0x4D
	PageDown = 0x4E
	RightArrow = 0x4F
	LeftArrow = 0x50
	DownArrow = 0x51
	UpArrow = 0x52
	NumpadLockandClear = 0x53
	NumpadDivide = 0x54
	NumpadMultiply = 0x55
	NumpadMinus = 0x56
	NumpadPlus = 0x57
	NumpadEnter = 0x58
	Numpad1 = 0x59
	Numpad2 = 0x5A
	Numpad3 = 0x5B
	Numpad4 = 0x5C
	Numpad5 = 0x5D
	Numpad6 = 0x5E
	Numpad7 = 0x5F
	Numpad8 = 0x60
	Numpad9 = 0x61
	Numpad0 = 0x62
	NumpadDot = 0x63
	NonUSBackslash = 0x64
	Application = 0x65
	Power = 0x66
	Equals = 0x67
	F13 = 0x68
	F14 = 0x69
	F15 = 0x6A
	F16 = 0x6B
	F17 = 0x6C
	F18 = 0x6D
	F19 = 0x6E
	F20 = 0x6F
	F21 = 0x70
	F22 = 0x71
	F23 = 0x72
	F24 = 0x73
	Execute = 0x74
	Help = 0x75
	Menu = 0x76
	Select = 0x77
	Stop = 0x78
	Again = 0x79
	Undo = 0x7A
	Cut = 0x7B
	Copy = 0x7C
	Paste = 0x7D
	Find = 0x7E
	Mute = 0x7F
	VolumeUp = 0x80
	VolumeDown = 0x81
	LockingCapsLock = 0x82
	LockingNumLock = 0x83
	LockingScrollLock = 0x84
	CommaSign = 0x85
	EqualSign = 0x86
	KeyboardInternational1 = 0x87
	KeyboardInternational2 = 0x88
	KeyboardInternational3 = 0x89
	KeyboardInternational4 = 0x8A
	KeyboardInternational5 = 0x8B
	KeyboardInternational6 = 0x8C
	KeyboardInternational7 = 0x8D
	KeyboardInternational8 = 0x8E
	KeyboardInternational9 = 0x8F
	KeyboardLANG1 = 0x90
	KeyboardLANG2 = 0x91
	KeyboardLANG3 = 0x92
	KeyboardLANG4 = 0x93
	KeyboardLANG5 = 0x94
	KeyboardLANG6 = 0x95
	KeyboardLANG7 = 0x96
	KeyboardLANG8 = 0x97
	KeyboardLANG9 = 0x98
	KeyboardAlternateErase = 0x99
	KeyboardSysReqAttention = 0x9A
	KeyboardCancel = 0x9B
	KeyboardClear = 0x9C
	KeyboardPrior = 0x9D
	KeyboardReturn = 0x9E
	KeyboardSeparator = 0x9F
	KeyboardOut = 0xA0
	KeyboardOper = 0xA1
	KeyboardClearAgain = 0xA2
	KeyboardCrSelProps = 0xA3
	KeyboardExSel = 0xA4
	Keypad00 = 0xB0
	Keypad000 = 0xB1
	ThousandsSeparator = 0xB2
	DecimalSeparator = 0xB3
	CurrencyUnit = 0xB4
	CurrencySubUnit34 = 0xB5
	KeypadLeftBracket = 0xB6
	KeypadRightBracket = 0xB7
	KeypadLeftBrace = 0xB8
	KeypadRightBrace = 0xB9
	KeypadTab = 0xBA
	Backspace = 0xBB
	KeypadA = 0xBC
	KeypadB = 0xBD
	KeypadC = 0xBE
	KeypadD = 0xBF
	KeypadE = 0xC0
	KeypadF = 0xC1
	Xox = 0xC2
	KeyXor = 0xC3
	KeyPercent = 0xC4
	KeyLessThan = 0xC5
	KeyMoreThan = 0xC6
	KeyAnd = 0xC7
	KeyAndAnd = 0xC8
	KeyOr = 0xC9
	KeyOrOr = 0xCA
	KeyColon = 0xCB
	KeyHash = 0xCC
	Space = 0xCD
	KeyAt = 0xCE
	KeyExclamation = 0xCF
	MemoryStore = 0xD0
	MemoryRecall = 0xD1
	MemoryClear = 0xD2
	MemoryAdd = 0xD3
	MemorySubtract = 0xD4
	MemoryMultiply = 0xD5
	MemoryDivide = 0xD6
	PlusMinus = 0xD7
	Clear = 0xD8
	ClearEntry = 0xD9
	Binary = 0xDA
	Octal = 0xDB
	Decimal = 0xDC
	KeypadHexadecimal = 0xDD
	KeyboardLeftControl = 0xE0
	KeyboardLeftShift = 0xE1
	KeyboardLeftAlt = 0xE2
	KeyboardLeftGUI = 0xE3
	KeyboardRightControl = 0xE4
	KeyboardRightShift = 0xE5
	KeyboardRightAlt = 0xE6
	KeyboardRightGUI = 0xE7


class LedPage(IntEnum):
	Undefined = 0x00
	NumLock = 0x01
	CapsLock = 0x02
	ScrollLock = 0x03
	Compose = 0x04
	Kana = 0x05
	Power = 0x06
	Shift = 0x07
	DoNotDisturb = 0x08
	Mute = 0x09
	ToneEnable = 0x0A
	HighCutFilter = 0x0B
	LowCutFilter = 0x0C
	EqualizerEnable = 0x0D
	SoundFieldOn = 0x0E
	SurroundOn = 0x0F
	Repeat = 0x10
	Stereo = 0x11
	SamplingRateDetect = 0x12
	Spinning = 0x13
	CAV = 0x14
	CLV = 0x15
	RecordingFormatDetect = 0x16
	OffHook = 0x17
	Ring = 0x18
	MessageWaiting = 0x19
	DataMode = 0x1A
	BatteryOperation = 0x1B
	BatteryOK = 0x1C
	BatteryLow = 0x1D
	Speaker = 0x1E
	HeadSet = 0x1F
	Hold = 0x20
	Microphone = 0x21
	Coverage = 0x22
	NightMode = 0x23
	SendCalls = 0x24
	CallPickup = 0x25
	Conference = 0x26
	StandBy = 0x27
	CameraOn = 0x28
	CameraOff = 0x29
	OnLine = 0x2A
	OffLine = 0x2B
	Busy = 0x2C
	Ready = 0x2D
	PaperOut = 0x2E
	PaperJam = 0x2F
	Remote = 0x30
	Forward = 0x31
	Reverse = 0x32
	Stop = 0x33
	Rewind = 0x34
	FastForward = 0x35
	Play = 0x36
	Pause = 0x37
	Record = 0x38
	Error = 0x39
	UsageSelectedIndicator = 0x3A
	UsageInUseIndicator = 0x3B
	UsageMultiModeIndicator = 0x3C
	IndicatorOn = 0x3D
	IndicatorFlash = 0x3E
	IndicatorSlowBlink = 0x3F
	IndicatorFastBlink = 0x40
	IndicatorOff = 0x41
	FlashOnTime = 0x42
	SlowBlinkOnTime = 0x43
	SlowBlinkOffTime = 0x44
	FastBlinkOnTime = 0x45
	FastBlinkOffTime = 0x46
	UsageIndicatorColor = 0x47
	IndicatorRed = 0x48
	IndicatorGreen = 0x49
	IndicatorAmber = 0x4A
	GenericIndicator = 0x4B
	SystemSuspend = 0x4C
	ExternalPowerConnected = 0x4D


class ButtonPage(IntEnum):
	NoButtonPressed = 0x00
	Button1 = 0x01
	Button2 = 0x02
	Button3 = 0x03
	Button4 = 0x04


class OrdinalPage(IntEnum):
	Instance1 = 0x01
	Instance2 = 0x02
	Instance3 = 0x03
	Instance4 = 0x04


class TelephonyDevicePage(IntEnum):
	Unassigned = 0x00
	Phone = 0x01
	AnsweringMachine = 0x02
	MessageControls = 0x03
	Handset = 0x04
	Headset = 0x05
	TelephonyKeyPad = 0x06
	ProgrammableButton = 0x07
	HookSwitch = 0x20
	Flash = 0x21
	Feature = 0x22
	Hold = 0x23
	Redial = 0x24
	Transfer = 0x25
	Drop = 0x26
	Park = 0x27
	ForwardCalls = 0x28
	AlternateFunction = 0x29
	Line = 0x2A
	SpeakerPhone = 0x2B
	Conference = 0x2C
	RingEnable = 0x2D
	RingSelect = 0x2E
	PhoneMute = 0x2F
	CallerID = 0x30
	Send = 0x31
	SpeedDial = 0x50
	StoreNumber = 0x51
	RecallNumber = 0x52
	PhoneDirectory = 0x53
	VoiceMail = 0x70
	ScreenCalls = 0x71
	DoNotDisturb = 0x72
	Message = 0x73
	AnswerOnOff = 0x74
	InsideDialTone = 0x90
	OutsideDialTone = 0x91
	InsideRingTone = 0x92
	OutsideRingTone = 0x93
	PriorityRingTone = 0x94
	InsideRingback = 0x95
	PriorityRingback = 0x96
	LineBusyTone = 0x97
	ReorderTone = 0x98
	CallWaitingTone = 0x99
	ConfirmationTone1 = 0x9A
	ConfirmationTone2 = 0x9B
	TonesOff = 0x9C
	OutsideRingback = 0x9D
	Ringer = 0x9E
	PhoneKey0 = 0xB0
	PhoneKey1 = 0xB1
	PhoneKey2 = 0xB2
	PhoneKey3 = 0xB3
	PhoneKey4 = 0xB4
	PhoneKey5 = 0xB5
	PhoneKey6 = 0xB6
	PhoneKey7 = 0xB7
	PhoneKey8 = 0xB8
	PhoneKey9 = 0xB9
	PhoneKeyStar = 0xBA
	PhoneKeyPound = 0xBB
	PhoneKeyA = 0xBC
	PhoneKeyB = 0xBD
	PhoneKeyC = 0xBE
	PhoneKeyD = 0xBF


class ConsumerPage(IntEnum):
	Unassigned = 0x00
	ConsumerControl = 0x01
	NumericKeyPad = 0x02
	ProgrammableButtons = 0x03
	Microphone = 0x04
	Headphone = 0x05
	GraphicEqualizer = 0x06
	Plus10 = 0x20
	Plus100 = 0x21
	AMPM = 0x22
	Power = 0x30
	Reset = 0x31
	Sleep = 0x32
	SleepAfter = 0x33
	SleepMode = 0x34
	Illumination = 0x35
	FunctionButtons = 0x36
	Menu = 0x40
	MenuPick = 0x41
	MenuUp = 0x42
	MenuDown = 0x43
	MenuLeft = 0x44
	MenuRight = 0x45
	MenuEscape = 0x46
	MenuValueIncrease = 0x47
	MenuValueDecrease = 0x48
	DataOnScreen = 0x60
	ClosedCaption = 0x61
	ClosedCaptionSelect = 0x62
	VCRTV = 0x63
	BroadcastMode = 0x64
	Snapshot = 0x65
	Still = 0x66
	Selection = 0x80
	AssignSelection = 0x81
	ModeStep = 0x82
	RecallLast = 0x83
	EnterChannel = 0x84
	OrderMovie = 0x85
	Channel = 0x86
	MediaSelection = 0x87
	MediaSelectComputer = 0x88
	MediaSelectTV = 0x89
	MediaSelectWWW = 0x8A
	MediaSelectDVD = 0x8B
	MediaSelectTelephone = 0x8C
	MediaSelectProgramGuide = 0x8D
	MediaSelectVideoPhone = 0x8E
	MediaSelectGames = 0x8F
	MediaSelectMessages = 0x90
	MediaSelectCD = 0x91
	MediaSelectVCR = 0x92
	MediaSelectTuner = 0x93
	Quit = 0x94
	Help = 0x95
	MediaSelectTape = 0x96
	MediaSelectCable = 0x97
	MediaSelectSatellite = 0x98
	MediaSelectSecurity = 0x99
	MediaSelectHome = 0x9A
	MediaSelectCall = 0x9B
	ChannelIncrement = 0x9C
	ChannelDecrement = 0x9D
	MediaSelectSAP = 0x9E
	VCRPlus = 0xA0
	Once = 0xA1
	Daily = 0xA2
	Weekly = 0xA3
	Monthly = 0xA4
	Play = 0xB0
	Pause = 0xB1
	Record = 0xB2
	FastForward = 0xB3
	Rewind = 0xB4
	ScanNextTrack = 0xB5
	ScanPreviousTrack = 0xB6
	Stop = 0xB7
	Eject = 0xB8
	RandomPlay = 0xB9
	SelectDisc = 0xBA
	EnterDisc = 0xBB
	Repeat = 0xBC
	Tracking = 0xBD
	TrackNormal = 0xBE
	SlowTracking = 0xBF
	FrameForward = 0xC0
	FrameBack = 0xC1
	Mark = 0xC2
	ClearMark = 0xC3
	RepeatFromMark = 0xC4
	ReturnToMark = 0xC5
	SearchMarkForward = 0xC6
	SearchMarkBackwards = 0xC7
	CounterReset = 0xC8
	ShowCounter = 0xC9
	TrackingIncrement = 0xCA
	TrackingDecrement = 0xCB
	StopEject = 0xCC
	PlayPause = 0xCD
	PlaySkip = 0xCE
	Volume = 0xE0
	Balance = 0xE1
	Mute = 0xE2
	Bass = 0xE3
	Treble = 0xE4
	BassBoost = 0xE5
	SurroundMode = 0xE6
	Loudness = 0xE7
	MPX = 0xE8
	VolumeIncrement = 0xE9
	VolumeDecrement = 0xEA
	SpeedSelect = 0xF0
	PlaybackSpeed = 0xF1
	StandardPlay = 0xF2
	LongPlay = 0xF3
	ExtendedPlay = 0xF4
	Slow = 0xF5
	FanEnable = 0x100
	FanSpeed = 0x101
	LightEnable = 0x102
	LightIlluminationLevel = 0x103
	ClimateControlEnable = 0x104
	RoomTemperature = 0x105
	SecurityEnable = 0x106
	FireAlarm = 0x107
	PoliceAlarm = 0x108
	Proximity = 0x109
	Motion = 0x10A
	DuressAlarm = 0x10B
	HoldupAlarm = 0x10C
	MedicalAlarm = 0x10D
	BalanceRight = 0x150
	BalanceLeft = 0x151
	BassIncrement = 0x152
	BassDecrement = 0x153
	TrebleIncrement = 0x154
	TrebleDecrement = 0x155
	SpeakerSystem = 0x160
	ChannelLeft = 0x161


class DigitizersPage(IntEnum):
	Undefined = 0x00
	Digitizer = 0x01
	Pen = 0x02
	LightPen = 0x03
	TouchScreen = 0x04
	TouchPad = 0x05
	WhiteBoard = 0x06
	CoordinateMeasuringMachine = 0x07
	Digitizer3D = 0x08
	StereoPlotter = 0x09
	ArticulatedArm = 0x0A
	Armature = 0x0B
	MultiplePointDigitizer = 0x0C
	FreeSpaceWand = 0x0D
	Stylus = 0x20
	Puck = 0x21
	Finger = 0x22
	TipPressure = 0x30
	BarrelPressure = 0x31
	InRange = 0x32
	Touch = 0x33
	Untouch = 0x34
	Tap = 0x35
	Quality = 0x36
	DataValid = 0x37
	TransducerIndex = 0x38
	TabletFunctionKeys = 0x39
	ProgramChangeKeys = 0x3A
	BatteryStrength = 0x3B
	Invert = 0x3C
	XTilt = 0x3D
	YTilt = 0x3E
	Azimuth = 0x3F
	Altitude = 0x40
	Twist = 0x41
	TipSwitch = 0x42
	SecondaryTipSwitch = 0x43
	BarrelSwitch = 0x44
	Eraser = 0x45
	TabletPick = 0x46


class AlphanumericDisplayPage(IntEnum):
	Undefined = 0x00
	AlphanumericDisplay = 0x01
	BitmappedDisplay = 0x02
	DisplayAttributesReport = 0x20
	ASCIICharacterSet = 0x21
	DataReadBack = 0x22
	FontReadBack = 0x23
	DisplayControlReport = 0x24
	ClearDisplay = 0x25
	DisplayEnable = 0x26
	ScreenSaverDelay = 0x27
	ScreenSaverEnable = 0x28
	VerticalScroll = 0x29
	HorizontalScroll = 0x2A
	CharacterReport = 0x2B
	DisplayData = 0x2C
	DisplayStatus = 0x2D
	StatNotReady = 0x2E
	StatReady = 0x2F
	ErrNotaloadablecharacter = 0x30
	ErrFontdatacannotberead = 0x31
	CursorPositionReport = 0x32
	Row = 0x33
	Column = 0x34
	Rows = 0x35
	Columns = 0x36
	CursorPixelPositioning = 0x37
	CursorMode = 0x38
	CursorEnable = 0x39
	CursorBlink = 0x3A
	FontReport = 0x3B
	FontData = 0x3C
	CharacterWidth = 0x3D
	CharacterHeight = 0x3E
	CharacterSpacingHorizontal = 0x3F
	CharacterSpacingVertical = 0x40
	UnicodeCharacterSet = 0x41
	Font7Segment = 0x42
	DirectMap7Segment = 0x43
	Font14Segment = 0x44
	DirectMap14Segment = 0x45
	DisplayBrightness = 0x46
	DisplayContrast = 0x47
	CharacterAttribute = 0x48
	AttributeReadback = 0x49
	AttributeData = 0x4A
	CharAttrEnhance = 0x4B
	CharAttrUnderline = 0x4C
	CharAttrBlink = 0x4D
	BitmapSizeX = 0x80
	BitmapSizeY = 0x81
	BitDepthFormat = 0x83
	DisplayOrientation = 0x84
	PaletteReport = 0x85
	PaletteDataSize = 0x86
	PaletteDataOffset = 0x87
	PaletteData = 0x88
	BlitReport = 0x8A
	BlitRectangleX1 = 0x8B
	BlitRectangleY1 = 0x8C
	BlitRectangleX2 = 0x8D
	BlitRectangleY2 = 0x8E
	BlitData = 0x8F
	SoftButton = 0x90
	SoftButtonID = 0x91
	SoftButtonSide = 0x92
	SoftButtonOffset1 = 0x93
	SoftButtonOffset2 = 0x94
	SoftButtonReport = 0x95


class SensorPage(IntEnum):
	SensorTypeCollection = 0x01
	BiometricHumanPresence = 0x11
	BiometricHumanProximity = 0x12
	BiometricHumanTouch = 0x13
	ElectricalCurrent = 0x22
	ElectricalPower = 0x23
	ElectricalVoltage = 0x26
	ElectricalPotentiometer = 0x27
	ElectricalFrequency = 0x28
	EnvironmentalAtmosphericPressure = 0x31
	EnvironmentalHumidity = 0x32
	EnvironmentalTemperature = 0x33
	LightAmbientLight = 0x41
	MechanicalBooleanSwitch = 0x61
	MechanicalBooleanSwitchArray = 0x62
	MechanicalMultivalueSwitch = 0x63
	MotionAccelerometer1D = 0x71
	MotionAccelerometer2D = 0x72
	MotionAccelerometer3D = 0x73
	MotionGyrometer1D = 0x74
	MotionGyrometer2D = 0x75
	MotionGyrometer3D = 0x76
	MotionMotionDetector = 0x77
	OrientationCompass1D = 0x81
	OrientationCompass3D = 0x83
	OrientationInclinometer1D = 0x84
	OrientationInclinometer2D = 0x85
	OrientationInclinometer3D = 0x86
	OrientationDistance1D = 0x87
	OrientationDistance2D = 0x88
	OrientationDistance3D = 0x89
	OrientationDeviceOrientation = 0x8A
	OtherCustom = 0xE1
	OtherGeneric = 0xE2


class MedicalInstrumentPage(IntEnum):
	Undefined = 0x00
	MedicalUltrasound = 0x01
	VCRAcquisition = 0x20
	FreezeThaw = 0x21
	ClipStore = 0x22
	Update = 0x23
	Next = 0x24
	Save = 0x25
	Print = 0x26
	MicrophoneEnable = 0x27
	Cine = 0x40
	TransmitPower = 0x41
	Volume = 0x42
	Focus = 0x43
	Depth = 0x44
	SoftStepPrimary = 0x60
	SoftStepSecondary = 0x61
	DepthGainCompensation = 0x70
	ZoomSelect = 0x80
	ZoomAdjust = 0x81
	SpectralDopplerModeSelect = 0x82
	SpectralDopplerAdjust = 0x83
	ColorDopplerModeSelect = 0x84
	ColorDopplerAdjust = 0x85
	MotionModeSelect = 0x86
	MotionModeAdjust = 0x87
	ModeSelect2D = 0x88
	ModeAdjust2D = 0x89
	SoftControlSelect = 0xA0
	SoftControlAdjust = 0xA1


class VendorPage(IntEnum):
	Usage1 = 0x01
	Usage2 = 0x02
	Usage3 = 0x03
	Usage4 = 0x04
	Usage5 = 0x05
	Usage6 = 0x06
	Usage7 = 0x07
	Usage8 = 0x08
	Usage9 = 0x09
	Usage10 = 0x0A
	Usage11 = 0x0B
	Usage12 = 0x0C


class UsagePage(IntEnum):
	GenericDesktopPage = 0x01
	SimulationControlsPage = 0x02
	VRControlsPage = 0x03
	SportControlsPage = 0x04
	GameControlsPage = 0x05
	GenericDeviceControlsPage = 0x06
	KeyboardKeypadPage = 0x07
	LedPage = 0x08
	ButtonPage = 0x09
	OrdinalPage = 0x0A
	TelephonyDevicePage = 0x0B
	ConsumerPage = 0x0C
	DigitizersPage = 0x0D
	AlphanumericDisplayPage = 0x14
	SensorPage = 0x20
	MedicalInstrumentPage = 0x40
	VendorPage = 0xFF00


page_to_enum = {
	UsagePage.GenericDesktopPage: GenericDesktopPage,
	UsagePage.SimulationControlsPage: SimulationControlsPage,
	UsagePage.VRControlsPage: VRControlsPage,
	UsagePage.SportControlsPage: SportControlsPage,
	UsagePage.GameControlsPage: GameControlsPage,
	UsagePage.GenericDeviceControlsPage: GenericDeviceControlsPage,
	UsagePage.KeyboardKeypadPage: KeyboardKeypadPage,
	UsagePage.LedPage: LedPage,
	UsagePage.ButtonPage: ButtonPage,
	UsagePage.OrdinalPage: OrdinalPage,
	UsagePage.TelephonyDevicePage: TelephonyDevicePage,
	UsagePage.ConsumerPage: ConsumerPage,
	UsagePage.DigitizersPage: DigitizersPage,
	UsagePage.AlphanumericDisplayPage: AlphanumericDisplayPage,
	UsagePage.SensorPage: SensorPage,
	UsagePage.MedicalInstrumentPage: MedicalInstrumentPage,
	UsagePage.VendorPage: VendorPage,
}


class SensorEvent(IntEnum):
	Event = 0x00
	SensorState = 0x01
	SensorEvent = 0x02


class HidSensorProperty(IntEnum):
	FriendlyName = 0x01
	PersistentUniqueID = 0x02
	MinimumReportInterval = 0x04
	SensorManufacturer = 0x05
	SensorModel = 0x06
	SensorSerialNumber = 0x07
	SensorDescription = 0x08
	SensorConnectionType = 0x09
	ReportInterval = 0x0E
	ChangeSensitivityAbsolute = 0x0F
	ChangeSensitivityrelativepercent = 0x11
	SensorAccuracy = 0x12
	SensorResolution = 0x13
	RangeMaximum = 0x14
	RangeMinimum = 0x15
	ReportingState = 0x16
	ResponseCurve = 0x18
	PowerState = 0x19


class MotionSensor(IntEnum):
	MotionState = 0x51
	Acceleration = 0x52
	AccelerationAxisX = 0x53
	AccelerationAxisY = 0x54
	AccelerationAxisZ = 0x55
	AngularVelocity = 0x56
	AngularVelocityXAxis = 0x57
	AngularVelocityYAxis = 0x58
	AngularVelocityZAxis = 0x59
	MotionIntensity = 0x5F


class OrientationSensor(IntEnum):
	Heading = 0x71
	HeadingCompensatedMagneticNorth = 0x75
	HeadingCompensatedTrueNorth = 0x76
	HeadingMagneticNorth = 0x77
	HeadingTrueNorth = 0x78
	Distance = 0x79
	DistanceXAxis = 0x7A
	DistanceYAxis = 0x7B
	DistanceZAxis = 0x7C
	DistanceOutOfRange = 0x7D
	Tilt = 0x7E
	TiltXAxis = 0x7F
	TiltYAxis = 0x80
	TiltZAxis = 0x81
	RotationMatrix = 0x82
	Quaternion = 0x83
	MagneticFlux = 0x84
	MagneticFluxXAxis = 0x85
	MagneticFluxYAxis = 0x86
	MagneticFluxZAxis = 0x87
	MagneticAccuracy = 0x88


class LightSensor(IntEnum):
	Illuminance = 0xD1
	ColorTemperature = 0xD2
	Chromaticity = 0xD3
	ChromaticityX = 0xD4
	ChromaticityY = 0xD5


class SensorDataField(IntEnum):
	Electrical = 0x00
	CapacitanceFarads = 0x01
	CurrentAmperes = 0x02
	ElectricalWatts = 0x03
	InductanceHenrys = 0x04
	ResistanceOhms = 0x05
	VoltageVolts = 0x06
	FrequencyHertz = 0x07
	PeriodMilliseconds = 0x08
	PercentOfRange = 0x09
	Custom = 0x40
	CustomUsage = 0x41
	CustomBooleanArray = 0x42
	CustomValue = 0x43
	CustomValue1 = 0x44
	CustomValue2 = 0x45
	CustomValue3 = 0x46
	CustomValue4 = 0x47
	CustomValue5 = 0x48
	CustomValue6 = 0x49
	CustomValue7 = 0x4A
	CustomValue8 = 0x4B
	CustomValue9 = 0x4C
	CustomValue10 = 0x4D
	CustomValue11 = 0x4E
	CustomValue12 = 0x4F
	CustomValue13 = 0x50
	CustomValue14 = 0x51
	CustomValue15 = 0x52
	CustomValue16 = 0x53
	CustomValue17 = 0x54
	CustomValue18 = 0x55
	CustomValue19 = 0x56
	CustomValue20 = 0x57
	CustomValue21 = 0x58
	CustomValue22 = 0x59
	CustomValue23 = 0x5A
	CustomValue24 = 0x5B
	CustomValue25 = 0x5C
	CustomValue26 = 0x5D
	CustomValue27 = 0x5E
	CustomValue28 = 0x5F


class SensorSelector(IntEnum):
	StateUnknown = 0x00
	StateNotAvailable = 0x01
	StateReady = 0x02
	StateNoData = 0x03
	StateInitializing = 0x04
	StateAccessDenied = 0x05
	StateError = 0x06
	EventUnknown = 0x10
	EventStateChanged = 0x11
	EventPropertyChanged = 0x12
	EventDataUpdated = 0x13
	EventPollResponse = 0x14
	EventChangeSensitivity = 0x15
	EventMaxReached = 0x16
	EventMinReached = 0x17
	EventHighThesholdCrossAbove = 0x18
	EventHighThresholdCrossBelow = 0x19
	EventLowThresholdCrossAbove = 0x1A
	EventLowThresholdCrossBelow = 0x1B
	EventZeroThresholdCrossAbove = 0x1C
	EventZeroThresholdCrossBelow = 0x1D
	EventPeriodExceeded = 0x1E
	EventFrequencyExceeded = 0x1F
	EventComplexTrigger = 0x20
	ConnectionTypePCIntegrated = 0x30
	ConnectionTypePCAttached = 0x31
	ConnectionTypePCExternal = 0x32
	PropertyReportingStateNoEvents = 0x40
	PropertyReportingStateAllEvents = 0x41
	PropertyReportingStateThresholdEvents = 0x42
	PropertyReportingStateNoEventsWake = 0x43
	PropertyReportingStateAllEventsWake = 0x44
	PropertyReportingStateThresholdEventsWake = 0x45
	PropertyPowerStateUndefined = 0x50
	PropertyPowerStateD0FullPower = 0x51
	PropertyPowerStateD1LowPower = 0x52
	PropertyPowerStateD2StandbyWithWake = 0x53
	PropertyPowerStateD3SleepWithWake = 0x54
	PropertyPowerStateD4PowerOff = 0x55
	DataFixQualityNoFix = 0x70
	DataFixQualityGps = 0x71
	DataFixQualityDgps = 0x72
	DataFixTypeNoFix = 0x80
	DataFixTypeGpsSpsModeFixValid = 0x81
	DataFixTypeDgpsSpsModeFixValid = 0x82
	DataFixTypeGpsPpsModeFixValid = 0x83
	DataFixTypeRealTimeKinematic = 0x84
	DataFixTypeFloatRtk = 0x85
	DataFixTypeEstimatedDeadReckoning = 0x86
	DataFixTypeManualInputMode = 0x87
	DataFixTypeSimulatorMode = 0x88
	DataGpsOpModeManual = 0x90
	DataGpsOpModeAutomatic = 0x91
	DataGpsSelModeAutonomous = 0xA0
	DataGpsSelModeDgps = 0xA1
	DataGpsSelModeEstimatedDeadReckoning = 0xA2
	DataGpsSelModeManualInput = 0xA3
	DataGpsSelModeSimulator = 0xA4
	DataGpsSelModeDataNotValid = 0xA5
	DataGpsStatusDataValid = 0xB0
	DataGpsStatusDataNotValid = 0xB1
	DesiredAccuracyDefault = 0x60
	DesiredAccuracyHigh = 0x61
	DesiredAccuracyMedium = 0x62
	DesiredAccuracyLow = 0x63
	GorpkKindCategory = 0xD0
	GorpkKindType = 0xD1
	GorpkKindEvent = 0xD2
	GorpkKindProperty = 0xD3
	GorpkKindDatafield = 0xD4


class UnitType(IntEnum):
	NoUnit = 0x00
	SILinear = 0x01
	SIRotation = 0x02
	EnglishLinear = 0x03
	EnglishRotation = 0x04


class Unit(IntEnum):
	Length = 0x01
	Mass = 0x02
	Time = 0x03
	Temperature = 0x04
	Current = 0x05
	LuminousIntensity = 0x06


class ModifierI2a(IntEnum):
	ChangeSensitivityAbsolute = 0x1
	Maximum = 0x2
	Minimum = 0x3
	Accuracy = 0x4
	Resolution = 0x5
	ThresholdHigh = 0x6
	ThresholdLow = 0x7
	CalibrationOffset = 0x8
	CalibrationMultiplier = 0x9
	ReportInterval = 0xA
	FrequencyMax = 0xB
	PeriodMax = 0xC
	ChangeSensitivityPercentOfRange = 0xD
	ChangeSensitivityPercentRelative = 0xE
	VendorOEM = 0xF


class ItemType(IntEnum):
	Constant = 0x1
	Data = 0x2


class ItemLength(IntEnum):
	Variable = 0x1
	Array = 0x2


class ItemBase(IntEnum):
	Relative = 0x1
	Absolute = 0x2



"""SC-Controller - Action Editor.

Also doubles as Menu Item Editor in some cases
"""

from __future__ import annotations

import importlib
import logging
import math

from gi.repository import GLib, Gtk

from scc.actions import Action, NoAction, RingAction, TriggerAction
from scc.constants import CUT, LINEAR, MINIMUM, ROUND, TRIGGER_CLICK, HapticPos, SCButtons
from scc.gui.ae import AEComponent
from scc.gui.controller_widget import GYROS, PADS, PRESSABLE, STICKS, TRIGGERS
from scc.gui.dwsnc import headerbar
from scc.gui.editor import Editor
from scc.gui.macro_editor import MacroEditor
from scc.gui.modeshift_editor import ModeshiftEditor
from scc.gui.parser import GuiActionParser, InvalidAction
from scc.gui.ring_editor import RingEditor
from scc.gui.simple_chooser import SimpleChooser
from scc.macros import Macro
from scc.modifiers import (
	BallModifier,
	ClickModifier,
	DeadzoneModifier,
	FeedbackModifier,
	ModeModifier,
	NameModifier,
	RotateInputModifier,
	SensitivityModifier,
	SmoothModifier,
)
from scc.profile import Profile
from scc.special_actions import OSDAction
from scc.tools import _, nameof

log = logging.getLogger("ActionEditor")


COMPONENTS = (  # List of known modules (components) in scc.gui.ae package
	"axis",
	"axis_action",
	"buttons",
	"custom",
	"dpad",
	"gesture",
	"gyro",
	"gyro_action",
	"per_axis",
	"special_action",
	"tilt",
	"trigger",
	# OSK-only components
	"osk_action",
	"osk_buttons",
)
XYZ = "XYZ"  # Sensitivity settings keys
AFP = ("Amplitude", "Frequency", "Period")  # Feedback settings keys
SMT = ("Level", "Weight", "Filter")  # Smoothing setting keys
DZN = ("Lower", "Upper")  # Deadzone settings key
FEEDBACK_SIDES = [HapticPos.LEFT, HapticPos.RIGHT, HapticPos.BOTH]
DEADZONE_MODES = [CUT, ROUND, LINEAR, MINIMUM]


class ActionEditor(Editor):
	GLADE = "action_editor.glade"
	ERROR_CSS = " #error {background-color:green; color:red;} "

	AEC_MENUITEM = -1

	MODE_TO_MODS = {
		# Specified which modifiers are compatibile with which editor mode.
		# That way, stuff like Rotation settings is not shown when editor
		# is used to edit menu actions.
		Action.AC_BUTTON: Action.MOD_OSD | Action.MOD_FEEDBACK,
		Action.AC_TRIGGER: Action.MOD_OSD | Action.MOD_SENSITIVITY | Action.MOD_FEEDBACK | Action.MOD_DEADZONE,
		Action.AC_STICK: Action.MOD_OSD
		| Action.MOD_CLICK
		| Action.MOD_DEADZONE
		| Action.MOD_ROTATE
		| Action.MOD_SENSITIVITY
		| Action.MOD_FEEDBACK
		| Action.MOD_SMOOTH,
		Action.AC_PAD: Action.MOD_OSD
		| Action.MOD_CLICK
		| Action.MOD_DEADZONE
		| Action.MOD_ROTATE
		| Action.MOD_SENSITIVITY
		| Action.MOD_FEEDBACK
		| Action.MOD_SMOOTH
		| Action.MOD_BALL,
		Action.AC_GYRO: Action.MOD_OSD
		| Action.MOD_SENSITIVITY
		| Action.MOD_SENS_Z
		| Action.MOD_DEADZONE
		| Action.MOD_FEEDBACK,
		Action.AC_OSK: 0,
		Action.AC_MENU: Action.MOD_OSD,
		AEC_MENUITEM: 0,
	}

	def __init__(self, app, callback) -> None:
		Editor.__init__(self)
		self.app = app
		self.id = None
		self.components = []                 # List of available components
		self.loaded_components = {}          # by class name
		self.c_buttons = {}                  # Component-to-button dict
		self.sens_widgets = []               # Sensitivity sliders, labels and 'clear' buttons
		self.feedback_widgets = []           # Feedback settings sliders, labels and 'clear' buttons, plus default value as last item
		self.smoothing_widgets = []          # Smoothing settings sliders, labels and 'clear' buttons, plus default value as last item
		self.deadzone_widgets = []           # Deadzone settings sliders, labels and 'clear' buttons, plus default value as last item
		self.sens = [1.0, 1.0, 1.0]          # Sensitivity slider values
		self.sens_defaults = [1.0, 1.0, 1.0] # Clear button clears to this
		self.feedback = [0.0, 0.0, 0.0]      # Feedback slider values, set later
		self.deadzone = [0, 0]               # Deadzone slider values, set later
		self.deadzone_mode = None            # None for 'disabled'
		self.feedback_position = None        # None for 'disabled'
		self.smoothing = None                # None for 'disabled'
		self.friction = -1                   # -1 for 'disabled'
		self.click = False                   # Click modifier value. None for disabled
		self.rotation_angle = 0              # RotateInputModifier angle
		self.osd = False                     # 'OSD enabled' value.
		self.first_page_allowed = False
		self.setup_widgets()
		self.load_components()
		self.ac_callback = callback  # This is different callback than ButtonChooser uses
		Editor.install_error_css()
		self._action = NoAction()
		self._replaced_action = None
		self._selected_component = None
		self._modifiers_enabled = True
		self._multiparams = [None] * 8
		self._mode = None
		self._recursing = False

	def setup_widgets(self):
		Editor.setup_widgets(self)
		headerbar(self.builder.get_object("header"))
		for i in (0, 1, 2):
			self.sens_widgets.append(
				(
					self.builder.get_object("sclSens%s" % (XYZ[i],)),
					self.builder.get_object("lblSens%s" % (XYZ[i],)),
					self.builder.get_object("btClearSens%s" % (XYZ[i],)),
					self.builder.get_object("cbSensInvert%s" % (XYZ[i],)),
				),
			)
		for key in AFP:
			i = AFP.index(key)
			self.feedback[i] = self.builder.get_object("sclF%s" % (key,)).get_value()
			self.feedback_widgets.append(
				(
					self.builder.get_object("sclF%s" % (key,)),
					self.builder.get_object("lblF%s" % (key,)),
					self.builder.get_object("btClearF%s" % (key,)),
					self.feedback[i],  # default value
				),
			)
		for key in SMT:
			i = SMT.index(key)
			self.smoothing_widgets.append(
				(
					self.builder.get_object("lblSmooth%s" % (key,)),
					self.builder.get_object("sclSmooth%s" % (key,)),
					self.builder.get_object("btClearSmooth%s" % (key,)),
					self.builder.get_object("sclSmooth%s" % (key,)).get_value(),
				),
			)
		for key in DZN:
			i = DZN.index(key)
			self.deadzone[i] = self.builder.get_object("sclDZ%s" % (key,)).get_value()
			self.deadzone_widgets.append(
				(
					self.builder.get_object("lblDZ%s" % (key,)),
					self.builder.get_object("sclDZ%s" % (key,)),
					self.builder.get_object("btClearDZ%s" % (key,)),
					self.deadzone[i],  # default value
				),
			)

		if self.app.osd_mode:
			self.builder.get_object("entName").set_sensitive(False)

	def load_components(self):
		"""Loads list of editor components"""
		# Import and load components
		for c in COMPONENTS:
			self.load_component(c)
		self._selected_component = None

	def load_component(self, class_name: str):
		"""Loads and adds new component to editor.

		Returns component instance.
		"""
		if class_name in self.loaded_components:
			return self.loaded_components[class_name]
		mod = importlib.import_module("scc.gui.ae.%s" % (class_name,))
		for x in mod.__all__:
			cls = getattr(mod, x)
			if isinstance(cls, type) and issubclass(cls, AEComponent):
				if cls is not AEComponent:
					instance = cls(self.app, self)
					self.loaded_components[class_name] = instance
					self.components.append(instance)
					return instance

	def on_Dialog_destroy(self, *a):
		cbPreview = self.builder.get_object("cbPreview")
		cbPreview.set_active(False)
		self.remove_added_widget()
		if self._selected_component is not None:
			self._selected_component.hidden()

	def on_Dialog_key_press_event(self, window, event):
		if self.app.osd_mode and event.keyval == 65471:
			self.on_btOK_clicked()

	def set_osd_enabled(self, value):
		"""Sets value of OSD modifier checkbox, without firing any more events."""
		self._recursing = True
		self.osd = value
		self.builder.get_object("cbOSD").set_active(value)
		self._recursing = False

	def show(self, transient_for):
		Editor.show(self, transient_for)

	def close(self):
		self.on_Dialog_destroy()
		Editor.close(self)

	def get_id(self):
		"""Returns ID of input that is being edited."""
		return self.id

	def on_link(self, link):
		parser = GuiActionParser()
		if link.startswith("quick://"):
			action = parser.restart(link[8:]).parse()
			self.reset_active_component()
			self.set_action(action, from_custom=True)
		elif link == "grab://trigger_button":

			def cb(action):
				action = TriggerAction(TRIGGER_CLICK, action)
				self.set_action(action, from_custom=True)
				self.force_page("trigger")

			b = SimpleChooser(self.app, "buttons", cb)
			b.set_title(_("Select Button"))
			b.hide_axes()
			b.show(self.window)
		elif link.startswith("page://"):

			def cb():
				self.force_page(link[7:])

			GLib.timeout_add(0.1, cb)
		elif link.startswith("advanced://"):
			exMore = self.builder.get_object("exMore")
			rvMore = self.builder.get_object("rvMore")
			ntbMore = self.builder.get_object("ntbMore")
			assert exMore.get_visible()
			exMore.set_expanded(True)
			rvMore.set_reveal_child(True)
			if "#" in link:
				link, name = link.split("#")
				self.blink_widget(name)
			ntbMore.set_current_page(int(link.split("/")[-1]))
		else:
			log.warning("Activated unknown link: %s", link)

	def on_action_type_changed(self, clicked_button):
		"""Called when user clicks on one of Action Type buttons."""
		# Prevent recurson
		if self._recursing:
			return
		self._recursing = True
		# Don't allow user to deactivate buttons - I'm using them as
		# radio button and you can't 'uncheck' radiobutton by clicking on it
		if not clicked_button.get_active():
			clicked_button.set_active(True)
			self._recursing = False
			return

		component = None
		for c in self.c_buttons:
			b = self.c_buttons[c]
			if clicked_button == b:
				component = c
			else:
				b.set_active(False)
		self._recursing = False

		stActionModes = self.builder.get_object("stActionModes")
		component.set_action(self._mode, self._action)
		if self._selected_component is not None:
			if self._selected_component != component:
				self._selected_component.hidden()
		self._selected_component = component
		self._selected_component.shown()
		stActionModes.set_visible_child(component.get_widget())

		stActionModes.show_all()

	def force_page(self, component, remove_rest: bool = False):
		"""Force action editor to display page with the specified component.

		If 'remove_rest' is True, removes all other pages.

		Returns 'component'
		"""
		if type(component) is str:
			component = self.load_component(component)
			return self.force_page(component, remove_rest)

		stActionModes = self.builder.get_object("stActionModes")
		component.load()
		if remove_rest:
			for c in stActionModes.get_children():
				if c != component:
					stActionModes.remove(c)

		if component.get_widget() not in stActionModes.get_children():
			stActionModes.add(component.get_widget())

		component.set_action(self._mode, self._action)
		if self._selected_component is not None:
			if self._selected_component != component:
				self._selected_component.hidden()
		self._selected_component = component
		self._selected_component.shown()
		stActionModes.set_visible_child(component.get_widget())
		stActionModes.show_all()

		return component

	def get_name(self):
		"""Returns action name as set in editor entry."""
		entName = self.builder.get_object("entName")
		return entName.get_text().strip(" \t")

	def get_current_page(self):
		"""Returns currently displayed page (component)."""
		return self._selected_component

	def _set_title(self):
		"""Copies title from text entry into action instance."""
		entName = self.builder.get_object("entName")
		name = entName.get_text().strip(" \t\r\n")
		if len(name) < 1:
			self._action.name = None
		elif not self._action:
			self._action = NameModifier(name, self._action)
		else:
			# print(">>>", "_set_title", self._action, entName)
			self._action.name = name

	def blink_widget(self, name, time=500):
		GROUPS = {"cbBallMode": ("cbBallMode", "lblFriction", "sclFriction", "btClearFriction")}

		def blink(widgets, count):
			count = count - 1
			for widget in widgets:
				widget.set_opacity(1.0 if count % 2 == 0 else 0.1)
			if count > 0:
				GLib.timeout_add(time, blink, widgets, count)

		if name in GROUPS:
			blink([self.builder.get_object(x) for x in GROUPS[name]], 7)
		else:
			blink([self.builder.get_object(name)], 7)

	def hide_modifiers(self):
		"""Hides (and disables) all modifiers"""
		self.set_modifiers_enabled(False)
		self.builder.get_object("exMore").set_visible(False)

	def hide_advanced_settings(self):
		"""Hides entire 'Advanced Settings' expander.
		"""
		self.builder.get_object("exMore").set_visible(False)
		self.builder.get_object("rvMore").set_visible(False)

	def hide_modeshift(self):
		"""Hides Mode Shift button.
		Used when displaying ActionEditor from ModeshiftEditor
		"""
		self.builder.get_object("btModeshift").set_visible(False)

	def hide_macro(self):
		"""Hides Macro button.
		Used when editing macro of pad/stick bindings.
		"""
		self.builder.get_object("btMacro").set_visible(False)

	def hide_ring(self):
		"""Hides Ring Bindings button.
		Used when editing anything but pad.
		"""
		self.builder.get_object("btInnerRing").set_visible(False)

	def hide_action_buttons(self):
		"""Hides action buttons, effectivelly disallowing user to change action type"""
		for x in ("lblActionType", "vbActionButtons"):
			self.builder.get_object(x).set_visible(False)
		self.hide_modeshift()
		self.hide_macro()
		self.hide_ring()

	def hide_action_str(self):
		"""Hides bottom part with action displayed as string"""
		self.builder.get_object("vbActionStr").set_visible(False)
		self.builder.get_object("grEditor").set_property("margin-bottom", 30)

	def hide_editor(self):
		"""Hides everything but action buttons and action name field"""
		self.builder.get_object("stActionModes").set_visible(False)
		self.hide_action_str()
		self.hide_modeshift()
		self.hide_macro()
		self.hide_ring()

	def hide_name(self):
		"""Hide (and clear) name field.

		Used when displaying ActionEditor from MacroEditor
		"""
		self.builder.get_object("lblName").set_visible(False)
		self.builder.get_object("entName").set_visible(False)
		self.builder.get_object("entName").set_text("")

	def hide_clear(self):
		"""Hide clear buttton."""
		self.builder.get_object("btClear").set_visible(False)

	def on_btClearRotation_clicked(self, *a):
		self.builder.get_object("sclRotation").set_value(0.0)

	def on_btClearSens_clicked(self, source, *a):
		i = 0
		for scale, label, button, checkbox in self.sens_widgets:
			if source == button:
				scale.set_value(self.sens_defaults[i])
				i += 1

	def on_btClearFeedback_clicked(self, source, *a):
		for scale, label, button, default in self.feedback_widgets:
			if source == button:
				scale.set_value(default)

	def on_btClearSmoothing_clicked(self, source, *a):
		for label, scale, button, default in self.smoothing_widgets:
			if source == button:
				scale.set_value(default)

	def on_btClearDeadzone_clicked(self, source, *a):
		for label, scale, button, default in self.deadzone_widgets:
			if source == button:
				scale.set_value(default)

	def on_btClear_clicked(self, *a):
		"""Handler for clear button"""
		action = NoAction()
		if self.ac_callback is not None:
			self.ac_callback(self.id, action)
		self.close()

	def on_btOK_clicked(self, *a):
		"""Handler for OK button"""
		if self.ac_callback is not None:
			self._set_title()
			if self._mode == ActionEditor.AEC_MENUITEM:
				self.ac_callback(self.id, self)
			else:
				a = self.generate_modifiers(self._action, self._selected_component.NAME == "custom")
				self.ac_callback(self.id, a)
				self.ac_callback = None
			if self._selected_component:
				self._selected_component.on_ok(a)
		self.close()

	def on_btModeshift_clicked(self, *a):
		"""Convert current action into modeshift and send it to ModeshiftEditor"""
		e = ModeshiftEditor(self.app, self.ac_callback)
		action = ModeModifier(self.generate_modifiers(self._action, self._selected_component.NAME == "custom"))
		e.set_input(self.id, action, mode=self._mode)
		self.send_added_widget(e)
		self.close()
		e.show(self.get_transient_for())

	def on_btMacro_clicked(self, *a):
		"""Convert current action into macro and send it to MacroEditor"""
		e = MacroEditor(self.app, self.ac_callback)
		action = Macro(self.generate_modifiers(self._action, self._selected_component.NAME == "custom"))
		e.set_input(self.id, action, mode=self._mode)
		self.send_added_widget(e)
		self.close()
		e.show(self.get_transient_for())

	def on_btInnerRing_clicked(self, *a):
		"""Convert current action into ring bindings and send it to RingEditor"""
		e = RingEditor(self.app, self.ac_callback)
		action = RingAction(self.generate_modifiers(self._action, self._selected_component.NAME == "custom"))
		e.set_input(self.id, action, mode=self._mode)
		self.send_added_widget(e)
		self.close()
		e.show(self.get_transient_for())

	def on_exMore_activate(self, ex, *a):
		rvMore = self.builder.get_object("rvMore")
		rvMore.set_reveal_child(not ex.get_expanded())

	def update_modifiers(self, *a):
		"""Called when sensitivity, feedback or other modifier setting changes."""
		if self._recursing:
			return
		cbRequireClick = self.builder.get_object("cbRequireClick")
		cbFeedbackSide = self.builder.get_object("cbFeedbackSide")
		cbFeedback = self.builder.get_object("cbFeedback")
		grFeedback = self.builder.get_object("grFeedback")
		cbDeadzone = self.builder.get_object("cbDeadzone")
		cbDeadzoneMode = self.builder.get_object("cbDeadzoneMode")
		cbSmoothing = self.builder.get_object("cbSmoothing")
		rvSmoothing = self.builder.get_object("rvSmoothing")
		sclRotation = self.builder.get_object("sclRotation")
		sclFriction = self.builder.get_object("sclFriction")
		cbBallMode = self.builder.get_object("cbBallMode")
		cbOSD = self.builder.get_object("cbOSD")
		set_action = False

		# Friction
		if not self.builder.get_object("cbBallMode").get_active():
			friction = -1
		elif sclFriction.get_value() == 0:
			friction = 0
		else:
			friction = (10.0 ** sclFriction.get_value()) / 1000.0
		if self.friction != friction:
			self.friction = friction
			set_action = True

		# Sensitivity
		for i in range(len(self.sens)):
			target = self.sens_widgets[i][0].get_value()
			if self.sens_widgets[i][3].get_active():
				target = -target
			if self.sens[i] != target:
				self.sens[i] = target
				set_action = True

		# Feedback
		if cbFeedback.get_active():
			feedback_position = FEEDBACK_SIDES[cbFeedbackSide.get_active()]
		else:
			feedback_position = None
		if self.feedback_position != feedback_position:
			self.feedback_position = feedback_position
			set_action = True

		for i in range(len(self.feedback)):
			if self.feedback[i] != self.feedback_widgets[i][0].get_value():
				self.feedback[i] = self.feedback_widgets[i][0].get_value()
				set_action = True

		for i in range(len(self.feedback)):
			if self.feedback[i] != self.feedback_widgets[i][0].get_value():
				self.feedback[i] = self.feedback_widgets[i][0].get_value()
				set_action = True

		# Deadzone
		mode = DEADZONE_MODES[cbDeadzoneMode.get_active()] if cbDeadzone.get_active() else None
		if self.deadzone_mode != mode:
			self.deadzone_mode = mode
			set_action = True

		for i in range(len(self.deadzone)):
			if self.deadzone[i] != self.deadzone_widgets[i][1].get_value():
				self.deadzone[i] = self.deadzone_widgets[i][1].get_value()
				set_action = True

		# Smoothing
		if cbSmoothing.get_active():
			smoothing = (
				int(self.smoothing_widgets[0][1].get_value()),
				self.smoothing_widgets[1][1].get_value(),
				int(self.smoothing_widgets[2][1].get_value()),
			)
		else:
			smoothing = None
		if self.smoothing != smoothing:
			self.smoothing = smoothing
			set_action = True

		# Rest
		if self.click is not None and cbRequireClick.get_active() != self.click:
			self.click = cbRequireClick.get_active()
			set_action = True

		if self.osd is not None and cbOSD.get_active() != self.osd:
			self.osd = cbOSD.get_active()
			set_action = True

		if self.rotation_angle != sclRotation.get_value():
			self.rotation_angle = sclRotation.get_value()
			set_action = True

		if set_action:
			self.set_action(self._action)
			self._selected_component.modifier_updated()

	def generate_modifiers(self, action, from_custom: bool = False):
		"""Return Action with all modifiers from UI applied."""
		if not self._modifiers_enabled and not from_custom:
			# Editing in custom action dialog, don't meddle with that
			return action

		if isinstance(action, ModeModifier):
			args = []
			for k in action.mods:
				if action.mods[k] is not None:
					args += [k, self.generate_modifiers(ActionEditor.strip_modifiers(action.mods[k]))]
			if action.default:
				args += [self.generate_modifiers(ActionEditor.strip_modifiers(action.default))]
			return ModeModifier(*args)

		cm = action.get_compatible_modifiers()

		if (cm & Action.MOD_BALL) != 0 and self.friction >= 0:
			action = BallModifier(round(self.friction, 3), action)

		if (cm & Action.MOD_SENSITIVITY) != 0:
			# Strip 1.0's from sensitivity values
			sens = [] + self.sens
			while len(sens) > 0 and sens[-1] == 1.0:
				sens = sens[0:-1]

			if len(sens) > 0:
				# Build arguments
				sens.append(action)
				# Create modifier
				action = SensitivityModifier(*sens)

		if (cm & Action.MOD_FEEDBACK) != 0:
			if self.feedback_position != None:
				# Strip defaults from feedback values
				feedback = [] + self.feedback
				while len(feedback) > 0 and feedback[-1] == self.feedback_widgets[len(feedback) - 1][-1]:
					feedback = feedback[0:-1]

				cbFeedbackSide = self.builder.get_object("cbFeedbackSide")
				cbFeedback = self.builder.get_object("cbFeedback")
				grFeedback = self.builder.get_object("grFeedback")
				if from_custom or (cbFeedback.get_active() and grFeedback.get_sensitive()):
					# Build FeedbackModifier arguments
					feedback = [FEEDBACK_SIDES[cbFeedbackSide.get_active()]] + feedback
					feedback += [action]
					# Create modifier
					action = FeedbackModifier(*feedback)

		if (cm & Action.MOD_SMOOTH) != 0:
			if self.smoothing != None:
				action = SmoothModifier(*(list(self.smoothing) + [action]))

		if (cm & Action.MOD_DEADZONE) != 0:
			if self.deadzone_mode is not None:
				action = DeadzoneModifier(self.deadzone_mode, self.deadzone[0], self.deadzone[1], action)

		if (cm & Action.MOD_ROTATE) != 0:
			if self.rotation_angle != 0.0:
				action = RotateInputModifier(self.rotation_angle, action)

		if (cm & Action.MOD_OSD) != 0:
			if self.osd:
				action = OSDAction(action)

		if (cm & Action.MOD_CLICK) != 0:
			if self.click:
				action = ClickModifier(action)

		return action

	@staticmethod
	def is_editable_modifier(action) -> bool:
		"""Return True if provided action is instance of modifier that ActionEditor can display and edit.

		Return False for everything else, even if it is instalce of Modifier subclass.
		"""
		if isinstance(
			action,
			(
				ClickModifier,
				SensitivityModifier,
				DeadzoneModifier,
				FeedbackModifier,
				RotateInputModifier,
				SmoothModifier,
				BallModifier,
			),
		):
			return True
		return bool(isinstance(action, OSDAction) and action.action is not None)

	@staticmethod
	def strip_modifiers(action):
		"""Return action stripped of all modifiers that are editable by editor."""
		while action:
			if ActionEditor.is_editable_modifier(action):
				action = action.action
			else:
				return action
		return action

	def load_modifiers(self, action, index: int = -1):
		"""Parse action for modifiers and update UI accordingly.

		Return action without parsed modifiers.
		"""
		cbRequireClick = self.builder.get_object("cbRequireClick")
		sclRotation = self.builder.get_object("sclRotation")
		cbOSD = self.builder.get_object("cbOSD")

		while ActionEditor.is_editable_modifier(action):
			if isinstance(action, RotateInputModifier):
				self.rotation_angle = action.angle
				action = action.action
			if isinstance(action, OSDAction):
				self.osd = True
				action = action.action
			if isinstance(action, ClickModifier):
				self.click = True
				action = action.action
			if isinstance(action, FeedbackModifier):
				self.feedback_position = action.haptic.get_position()
				self.feedback[0] = action.haptic.get_amplitude()
				self.feedback[1] = action.haptic.get_frequency()
				self.feedback[2] = action.haptic.get_period()
				action = action.action
			if isinstance(action, SmoothModifier):
				self.smoothing = (action.level, action.multiplier, action.filter)
				action = action.action
			if isinstance(action, DeadzoneModifier):
				self.deadzone_mode = action.mode
				self.deadzone[0] = action.lower
				self.deadzone[1] = action.upper
				action = action.action
			if isinstance(action, SensitivityModifier):
				if index < 0:
					for i in range(len(self.sens)):
						self.sens[i] = action.speeds[i]
				else:
					self.sens[index] = action.speeds[0]
				action = action.action
			if isinstance(action, BallModifier):
				self.friction = action.friction
				action = action.action

		self._recursing = True
		cbRequireClick.set_active(self.click)
		cbOSD.set_active(self.osd)
		sclRotation.set_value(self.rotation_angle)
		for i in range(len(self.sens)):
			self.sens_widgets[i][3].set_active(self.sens[i] < 0)
			self.sens_widgets[i][0].set_value(abs(self.sens[i]))
		# Feedback
		cbFeedbackSide = self.builder.get_object("cbFeedbackSide")
		lblFeedbackSide = self.builder.get_object("lblFeedbackSide")
		if self.feedback_position != None:
			cbFeedback = self.builder.get_object("cbFeedback")
			cbFeedbackSide.set_active(FEEDBACK_SIDES.index(self.feedback_position))
			cbFeedback.set_active(True)
			for i in range(len(self.feedback)):
				self.feedback_widgets[i][0].set_value(self.feedback[i])
		for grp in self.feedback_widgets:
			for w in grp[0:-1]:
				w.set_sensitive(self.feedback_position is not None)
		lblFeedbackSide.set_sensitive(self.feedback_position is not None)
		cbFeedbackSide.set_sensitive(self.feedback_position is not None)

		# Smoothing
		cbSmoothing = self.builder.get_object("cbSmoothing")
		if self.smoothing:
			cbSmoothing.set_active(True)
			for i in range(len(self.smoothing_widgets)):
				self.smoothing_widgets[i][1].set_value(self.smoothing[i])
		for grp in self.smoothing_widgets:
			for w in grp[0:-1]:
				w.set_sensitive(cbSmoothing.get_active())

		# Ball
		sclFriction = self.builder.get_object("sclFriction")
		cbBallMode = self.builder.get_object("cbBallMode")
		if self.friction < 0:
			cbBallMode.set_active(False)
		elif self.friction == 0:
			cbBallMode.set_active(True)
			sclFriction.set_value(0)
		else:
			cbBallMode.set_active(True)
			sclFriction.set_value(math.log(self.friction * 1000.0, 10))

		# Deadzone
		cbDeadzoneMode = self.builder.get_object("cbDeadzoneMode")
		lblDeadzoneMode = self.builder.get_object("lblDeadzoneMode")
		if self.deadzone_mode is not None:
			cbDeadzone = self.builder.get_object("cbDeadzone")
			cbDeadzone.set_active(True)
			cbDeadzoneMode.set_active(DEADZONE_MODES.index(self.deadzone_mode))
			for i in range(len(self.deadzone)):
				self.deadzone_widgets[i][1].set_value(self.deadzone[i])

		for grp in self.deadzone_widgets:
			for w in grp[0:-1]:
				w.set_sensitive(self.deadzone_mode is not None)
		lblDeadzoneMode.set_sensitive(self.deadzone_mode is not None)
		cbDeadzoneMode.set_sensitive(self.deadzone_mode is not None)

		self._recursing = False

		return action

	def allow_first_page(self) -> None:
		"""Allow the first page to be used."""
		self.first_page_allowed = True

	def reset_active_component(self) -> None:
		"""Forget what component was selected so next call to set_action selects new one."""
		self._selected_component = None

	def set_action(self, action, from_custom: bool = False) -> None:
		"""Update Action field(s) on bottom and recolors apropriate image area, if such area exists."""
		entAction = self.builder.get_object("entAction")
		cbPreview = self.builder.get_object("cbPreview")
		btOK = self.builder.get_object("btOK")

		# Load modifiers and update UI if needed
		action = self.load_modifiers(action)

		# Check for InvalidAction and display error message if found
		if isinstance(action, InvalidAction):
			btOK.set_sensitive(False)
			entAction.set_name("error")
			entAction.set_text(str(action.error))
		else:
			entAction.set_name("entAction")
			btOK.set_sensitive(True)
			self._action = action
			action = self.generate_modifiers(action, from_custom)

			if hasattr(action, "string") and "\n" not in action.string:
				# Stuff generated by my special parser
				entAction.set_text(action.string)
			else:
				# Actions generated elsewhere
				entAction.set_text(action.to_string())
			self.enable_modifiers(self._action)
			self.enable_preview(self._action)

		# Send changed action into selected component
		if self._selected_component is None:
			for component in reversed(sorted(self.components, key=lambda a: a.PRIORITY)):
				if (component.CTXS & self._mode) != 0:
					if component.handles(self._mode, ActionEditor.strip_modifiers(action)):
						self._selected_component = component
						break
			if isinstance(action, InvalidAction):
				c = self.load_component("custom")
				if c in self.components and (self._mode & c.CTXS) != 0:
					self._selected_component = c
			elif not action and self.first_page_allowed:
				self._selected_component = self.load_component("first_page")
				stActionModes = self.builder.get_object("stActionModes")
				self._selected_component.load()
				self._selected_component.shown()
				stActionModes.add(self._selected_component.get_widget())
				stActionModes.set_visible_child(self._selected_component.get_widget())
			if self._selected_component:
				if self._selected_component in self.c_buttons:
					self.c_buttons[self._selected_component].set_active(True)
			if isinstance(action, InvalidAction):
				self._selected_component.set_action(self._mode, action)
		elif not self._selected_component.handles(self._mode, ActionEditor.strip_modifiers(action)):
			log.warning("selected_component no longer handles edited action")
			log.warning(self._selected_component)
			log.warning(ActionEditor.strip_modifiers(action).to_string())
			log.warning("(%s)", action.to_string())

		if cbPreview.get_sensitive() and cbPreview.get_active():
			self.apply_preview(action)

	def apply_preview(self, action):
		if self._replaced_action is None:
			self._replaced_action = self.ac_callback(self.id, action, mark_changed=False)
		else:
			self.ac_callback(self.id, action, mark_changed=False)

	def on_cbPreview_toggled(self, cb):
		if cb.get_active():
			a = self.generate_modifiers(self._action, self._selected_component.NAME == "custom")
			self.apply_preview(a)
		elif self._replaced_action is not None:
			if self.ac_callback:
				# Is None if OK button handler was executed
				self.ac_callback(self.id, self._replaced_action, mark_changed=False)
			self._replaced_action = None

	def enable_preview(self, action):
		"""Enables or disables and hides 'preview immediately' option, based on
		if currently selected action supports it.
		"""
		cbPreview = self.builder.get_object("cbPreview")

		enabled = action.strip().get_previewable()
		cbPreview.set_sensitive(enabled)

	def enable_modifiers(self, action):
		"""Enables or disables and hides modifier settings according to what
		is applicable for specified action AND what's allowed for current
		editor mode.

		Uses value returned by action.get_compatible_modifiers.
		"""
		cm = action.get_compatible_modifiers() & ActionEditor.MODE_TO_MODS[self._mode]

		# Feedback
		grFeedback = self.builder.get_object("grFeedback")
		grFeedback.set_sensitive((cm & Action.MOD_FEEDBACK) != 0)

		# Smoothing
		grSmoothing = self.builder.get_object("grSmoothing")
		grSmoothing.set_sensitive((cm & Action.MOD_SMOOTH) != 0)

		# Deadzone
		grDeadzone = self.builder.get_object("grDeadzone")
		grDeadzone.set_sensitive((cm & Action.MOD_DEADZONE) != 0)

		# Sensitivity
		grSensitivity = self.builder.get_object("grSensitivity")
		grSensitivity.set_sensitive((cm & Action.MOD_SENSITIVITY) != 0)
		for w in self.sens_widgets[2]:
			w.set_visible((cm & Action.MOD_SENS_Z) != 0)

		# Rotation
		for w in ("lblRotationHeader", "lblRotation", "sclRotation", "btClearRotation"):
			self.builder.get_object(w).set_sensitive((cm & Action.MOD_ROTATE) != 0)

		# Click
		cbRequireClick = self.builder.get_object("cbRequireClick")
		cbRequireClick.set_sensitive((cm & Action.MOD_CLICK) != 0)

		# Ball
		cbBallMode = self.builder.get_object("cbBallMode")
		cbBallMode.set_sensitive((cm & Action.MOD_BALL) != 0)
		for w in ("sclFriction", "lblFriction", "btClearFriction"):
			self.builder.get_object(w).set_sensitive(cbBallMode.get_active() and ((cm & Action.MOD_BALL) != 0))
		if cm & Action.MOD_BALL == 0:
			self.builder.get_object("cbBallMode").set_active(False)

		# OSD
		cbOSD = self.builder.get_object("cbOSD")
		cbOSD.set_sensitive(cm & Action.MOD_OSD != 0)

	def set_sensitivity(self, x: float, y: float = 1.0, z: float = 1.0) -> None:
		"""Sets sensitivity for edited action."""
		self._recursing = True
		xyz = [x, y, z]
		for i in range(len(self.sens)):
			self.sens[i] = xyz[i]
			self.sens_widgets[i][3].set_active(self.sens[i] < 0)
			self.sens_widgets[i][0].set_value(abs(self.sens[i]))
		self._recursing = False
		self.set_action(self._action)
		self._selected_component.modifier_updated()

	def get_sensitivity(self) -> list[float]:
		"""Returns sensitivity currently set in editor."""
		print(f"CURRENT SENS: {self.sens}")
		return self.sens

	def set_default_sensitivity(self, x: float, y: float = 1.0, z: float = 1.0):
		"""Sets default sensitivity values and, if sensitivity is currently set to defaults, updates it to these values."""
		xyz = x, y, z
		update = False
		self._recursing = True
		for i in (0, 1, 2):
			if self.sens[i] == self.sens_defaults[i]:
				self.sens[i] = xyz[i]
				self.sens_widgets[i][0].set_value(xyz[i])
				update = True
			self.sens_defaults[i] = xyz[i]
		self._recursing = False
		if update:
			self.update_modifiers()

	def get_mode(self):
		return self._mode

	def _set_mode(self, action, mode):
		"""Common part of editor setup"""
		self._mode = mode
		# Clear pages and 'action type' buttons
		entName = self.builder.get_object("entName")
		vbActionButtons = self.builder.get_object("vbActionButtons")
		stActionModes = self.builder.get_object("stActionModes")

		# Go throgh list of components and display buttons that are usable
		# with this mode
		self.c_buttons = {}
		for component in reversed(sorted(self.components, key=lambda a: a.PRIORITY)):
			if (mode & component.CTXS) != 0:
				b = Gtk.ToggleButton.new_with_label(component.get_button_title())
				vbActionButtons.pack_start(b, True, True, 2)
				b.connect("toggled", self.on_action_type_changed)
				self.c_buttons[component] = b

				component.load()
				if component.get_widget() not in stActionModes.get_children():
					stActionModes.add(component.get_widget())

		if action.name is None:
			entName.set_text("")
		else:
			entName.set_text(action.name)
		if vbActionButtons.get_visible():
			vbActionButtons.show_all()

	def on_sclFFrequency_format_value(self, scale, value):
		if value == 1:
			# Special case
			return " %0.2fHz" % (1.0 / value,)
		return "%0.2fmHz" % (100.0 / value,)

	def on_sclFriction_format_value(self, scale, value):
		if value <= 0:
			return "%0.3f" % (0,)
		if value >= 6:
			return "%0.3f" % (1000.00,)
		return "%0.3f" % ((10.0**value) / 1000.0)

	def on_btClearFriction_clicked(self, *a):
		sclFriction = self.builder.get_object("sclFriction")
		sclFriction.set_value(math.log(10 * 1000.0, 10))

	def set_input(self, id, action, mode=None):
		"""Setups action editor for editing specified input.

		Mode (buttton/axis/trigger...) is either provided or chosen based on id.
		Also sets title, but that can be overriden by calling set_title after.
		"""
		self.id = id
		if id in SCButtons.__members__.values() or mode in (Action.AC_MENU, Action.AC_BUTTON):
			if id in PRESSABLE:
				self.set_title(_("%s Press") % (nameof(id),))
			elif id in SCButtons.__members__.values():
				self.set_title(
					nameof(id),
				)
			self._set_mode(action, mode or Action.AC_BUTTON)
			self.hide_modifiers()
			self.set_action(action)
		elif id in TRIGGERS:
			self.set_title(_("%s Trigger") % (id,))
			self._set_mode(action, mode or Action.AC_TRIGGER)
			self.set_action(action)
			self.hide_macro()
			self.hide_ring()
		elif id in STICKS:
			if id == Profile.DPAD:
				self.set_title(_("DPAD"))
			elif id == Profile.RSTICK:
				self.set_title(_("Right Stick"))
			elif id == Profile.STICK:
				self.set_title(_("Stick"))
			else:
				raise ValueError(f"unknown id {id}")
			self._set_mode(action, mode or Action.AC_STICK)
			self.set_action(action)
			self.hide_macro()
			self.id = id
		elif id in GYROS:
			self.set_title(_("Gyro"))
			self._set_mode(action, mode or Action.AC_GYRO)
			self.set_action(action)
			self.hide_modeshift()
			self.hide_macro()
			self.hide_ring()
			self.id = Profile.GYRO
		elif id in PADS:
			self._set_mode(action, mode or Action.AC_PAD)
			self.set_action(action)
			self.hide_macro()
			if id == Profile.LPAD:
				self.set_title(_("Left Pad"))
			elif id == Profile.RPAD:
				self.set_title(_("Right Pad"))
			else:
				self.set_title(_("Touch Pad"))
		if mode == Action.AC_OSK:
			self.hide_name()
			self.hide_modeshift()
			self.hide_macro()
			self.hide_ring()
			# self.hide_rotation()
		elif mode == Action.AC_MENU:
			self.hide_modeshift()
			self.hide_macro()
			self.hide_ring()

	def set_menu_item(self, item, title_for_name_label: bool | None = None):
		"""Setups action editor in way that allows editing only action name.

		In this mode, callback is called with editor instance instead of
		generated action as 2nd argument.
		"""
		entName = self.builder.get_object("entName")
		self._mode = ActionEditor.AEC_MENUITEM
		if hasattr(item, "label") and item.label:
			entName.set_text(item.label)
		else:
			entName.set_text("")
		# self.hide_osd()
		self.hide_action_buttons()
		self.hide_modifiers()
		self.set_action(NoAction())
		self.id = item.id
		if title_for_name_label:
			self.builder.get_object("lblName").set_label(title_for_name_label)

	def set_modifiers_enabled(self, enabled: bool) -> None:
		exMore = self.builder.get_object("exMore")
		rvMore = self.builder.get_object("rvMore")
		if self._modifiers_enabled != enabled and not enabled:
			exMore.set_expanded(False)
			rvMore.set_reveal_child(False)
		exMore.set_sensitive(enabled)
		self._modifiers_enabled = enabled



#!/usr/bin/env python3
"""SC-Controller - Controller Registration Constants

Just huge chunk of constants put aside to make impotant code more readable
"""

from scc.constants import LEFT, RIGHT, STICK, SCButtons

X = 0
Y = 1

AXIS_ORDER = (
	("stick_x", X),
	("stick_y", Y),
	("rpad_x", X),
	("rpad_y", Y),
	("lpad_x", X),
	("lpad_y", Y),
	("ltrig", X),  # index 6
	("rtrig", X),
)

STICK_PAD_AREAS = {
	# Numbers here are indexes to AXIS_ORDER tuple
	"STICK": (STICK, (0, 1)),
	"RPAD": (RIGHT, (2, 3)),
	"LPAD": (LEFT, (4, 5)),
}

TRIGGER_AREAS = {
	# Numbers here are indexes to AXIS_ORDER tuple
	"LT": 6,
	"RT": 7,
}

AXIS_TO_BUTTON = {
	# Maps stick and dpad axes to their respective "pressed" button
	"stick_x": SCButtons.STICKPRESS,
	"stick_y": SCButtons.STICKPRESS,
	"rpad_x": SCButtons.RPAD,
	"rpad_y": SCButtons.RPAD,
	"lpad_x": SCButtons.LPAD,
	"lpad_y": SCButtons.LPAD,
}

SDL_TO_SCC_NAMES = {
	"guide": "C",
	"leftstick": "STICKPRESS",
	"rightstick": "RPAD",
	"leftshoulder": "LB",
	"rightshoulder": "RB",
}

SDL_AXES = (
	# This tuple has to use same order as AXIS_ORDER
	"leftx",
	"lefty",
	"rightx",
	"righty",
	"dpadx",
	"dpady",
	"lefttrigger",
	"righttrigger",
)


SDL_DPAD = {
	# Numbers here are indexes to AXIS_ORDER tuple
	# Booleans here are True for positive movements (down/right) and
	# False for negative (up/left)
	"dpdown": (5, True),
	"dpleft": (4, False),
	"dpright": (4, True),
	"dpup": (5, False),
}



"""SC Controller - BindingEditor

Base class for main application window and OSD Keyboard bindings editor.
"""

from scc.actions import NoAction
from scc.constants import LEFT, RIGHT, SCButtons
from scc.gui.action_editor import ActionEditor
from scc.gui.ae.buttons import is_button_repeat, is_button_togle
from scc.gui.ae.gyro_action import is_gyro_enable
from scc.gui.controller_widget import (
	BUTTONS,
	GYROS,
	PADS,
	PRESSABLE,
	STICKS,
	TRIGGERS,
	ControllerButton,
	ControllerGyro,
	ControllerPad,
	ControllerStick,
	ControllerTrigger,
)
from scc.gui.macro_editor import MacroEditor
from scc.gui.modeshift_editor import ModeshiftEditor
from scc.gui.ring_editor import RingEditor
from scc.macros import Macro, Type
from scc.modifiers import DoubleclickModifier, FeedbackModifier, HoldModifier, ModeModifier, SensitivityModifier
from scc.profile import Profile
from scc.tools import _


class BindingEditor:
	def __init__(self, app):
		self.button_widgets = {}
		self.app = app

	def create_binding_buttons(self, use_icons=True, enable_press=True):
		"""Create ControllerWidget instances for available Gtk.Buttons defined in glade file."""
		for b in BUTTONS:
			w = self.builder.get_object("bt" + b.name)
			if w:
				self.button_widgets[b] = ControllerButton(self, b, use_icons, w)
		for b in TRIGGERS:
			w = self.builder.get_object("bt" + b)
			if w:
				self.button_widgets[b] = ControllerTrigger(self, b, use_icons, w)
		for b in PADS:
			w = self.builder.get_object("bt" + b)
			if w:
				self.button_widgets[b] = ControllerPad(self, b, use_icons, enable_press, w)
		for b in STICKS:
			w = self.builder.get_object("bt" + b)
			if w:
				e = False if b == Profile.DPAD else enable_press
				self.button_widgets[b] = ControllerStick(self, b, use_icons, e, w)
		w = self.builder.get_object("btSTICKPRESS")
		if w:
			self.button_widgets[SCButtons.STICKPRESS] = ControllerButton(self, SCButtons.STICKPRESS, use_icons, w)
		for b in GYROS:
			w = self.builder.get_object("bt" + b)
			if w:
				self.button_widgets[b] = ControllerGyro(self, b, use_icons, w)

	def on_action_chosen(self, id, action, mark_changed=True):
		"""Callback called when action editting is finished in editor.
		Should return None or action being replaced.
		"""
		raise TypeError("Non-overriden on_action_chosen")

	def set_action(self, profile, id, action):
		"""Stores action in profile.
		Returns formely stored action.
		"""
		before = NoAction()
		if id == SCButtons.STICKPRESS and Profile.STICK in self.button_widgets:
			before, profile.buttons[id] = profile.buttons[id], action
			self.button_widgets[Profile.STICK].update()
		elif id == SCButtons.CPADPRESS and Profile.CPAD in self.button_widgets:
			before, profile.buttons[id] = profile.buttons[id], action
			self.button_widgets[Profile.CPAD].update()
		elif id in PRESSABLE:
			before, profile.buttons[id] = profile.buttons[id], action
			self.button_widgets[id.name].update()
		elif id in BUTTONS:
			before, profile.buttons[id] = profile.buttons[id], action
			self.button_widgets[id].update()
		elif id in TRIGGERS:
			# TODO: Use LT and RT in profile as well
			side = LEFT if id == "LT" else RIGHT
			before, profile.triggers[side] = profile.triggers[side], action
			self.button_widgets[id].update()
		elif id in GYROS:
			before, profile.gyro = profile.gyro, action
			self.button_widgets[id].update()
		elif id in STICKS + PADS:
			if id == Profile.STICK:
				before, profile.stick = profile.stick, action
			elif id == Profile.RSTICK:
				before, profile.rstick = profile.rstick, action
			elif id == Profile.DPAD:
				before, profile.pads[Profile.DPAD] = profile.pads[Profile.DPAD], action
			elif id == Profile.LPAD:
				before, profile.pads[Profile.LEFT] = profile.pads[Profile.LEFT], action
			elif id == Profile.RPAD:
				before, profile.pads[Profile.RIGHT] = profile.pads[Profile.RIGHT], action
			elif id == Profile.CPAD:
				before, profile.pads[Profile.CPAD] = profile.pads[Profile.CPAD], action
			else:
				raise ValueError("unknown id %s" % (id,))
			self.button_widgets[id].update()
		return before

	def get_action(self, profile, id):
		"""Returns action for specified id.
		Returns None if id is not known.
		"""
		before = NoAction()
		if id in BUTTONS or id in PRESSABLE:
			return profile.buttons[id]
		if id in TRIGGERS:
			# TODO: Use LT and RT in profile as well
			side = LEFT if id == "LT" else RIGHT
			return profile.triggers[side]
		if id in GYROS:
			return profile.gyro
		if id in STICKS + PADS:
			if id == Profile.STICK:
				return profile.stick
			if id == Profile.RSTICK:
				return profile.rstick
			if id in Profile.DPAD:
				return profile.pads[Profile.DPAD]
			if id == Profile.LPAD:
				return profile.pads[Profile.LEFT]
			if id == Profile.RPAD:
				return profile.pads[Profile.RIGHT]
			if id == Profile.CPAD:
				return profile.pads[Profile.CPAD]
			raise ValueError("unknown id %s" % (id,))
		return None

	def choose_editor(self, action, title, id=None):
		"""Chooses apropripate Editor instance for edited action"""
		if isinstance(action, SensitivityModifier):
			action = action.action
		if isinstance(action, FeedbackModifier):
			action = action.action
		if id in GYROS:
			e = ActionEditor(self.app, self.on_action_chosen)
			e.set_title(title)
		elif isinstance(action, (ModeModifier, DoubleclickModifier, HoldModifier)) and not is_gyro_enable(action):
			e = ModeshiftEditor(self.app, self.on_action_chosen)
			e.set_title(_("Mode Shift for %s") % (title,))
		elif RingEditor.is_ring_action(action):
			e = RingEditor(self.app, self.on_action_chosen)
			e.set_title(title)
		elif isinstance(action, Type):
			# Type is subclass of Macro
			e = ActionEditor(self.app, self.on_action_chosen)
			e.set_title(title)
		elif isinstance(action, Macro) and not (is_button_togle(action) or is_button_repeat(action)):
			e = MacroEditor(self.app, self.on_action_chosen)
			e.set_title(_("Macro for %s") % (title,))
		else:
			e = ActionEditor(self.app, self.on_action_chosen)
			e.set_title(title)
		return e

	def hilight(self, button):
		"""Hilights button on image. Overriden by app."""

	def show_editor(self, id):
		raise TypeError("show_editor not overriden")



"""SC-Controller - OSD Menu.

Display menu that user can navigate through and prints chosen item id to stdout
"""

import logging
import os
import sys
from math import sqrt

from gi.repository import Gdk, GdkPixbuf, GdkX11, Gio, GLib, Gtk

from scc.config import Config
from scc.constants import DEFAULT, LEFT, RIGHT, SAME, STICK, STICK_PAD_MAX, ControllerFlags, SCButtons
from scc.gui.daemon_manager import DaemonManager
from scc.lib import xwrappers as X
from scc.menu_data import MenuData, Separator, Submenu
from scc.osd import OSDWindow, StickController, menu_generators
from scc.paths import get_share_path
from scc.tools import _, circle_to_square, clamp, find_icon, find_menu
from scc.x11 import autoswitcher

log = logging.getLogger("osd.menu")


menu_generators.register_menu_generators()
autoswitcher.register_menu_generators()

class Menu(OSDWindow):
	EPILOG = """Exit codes:
   0  - clean exit, user selected option
  -1  - clean exit, user canceled menu
  -2  - clean exit, menu closed from callback method
   1  - error, invalid arguments
   2  - error, failed to access sc-daemon, sc-daemon reported error or died while menu is displayed.
   3  - erorr, failed to lock input stick, pad or button(s)
	"""
	SUBMENU_OFFSET = 50
	PREFER_BW_ICONS = True

	def __init__(self, cls="osd-menu", layer=None):
		OSDWindow.__init__(self, cls, layer)
		self.daemon = None
		self.config = None
		self.feedback = None
		self.controller = None
		if isinstance(Gdk.Display.get_default(), GdkX11.X11Display):
			self.xdisplay = X.Display(hash(GdkX11.x11_get_default_xdisplay()))  # Magic
		else:
			self.xdisplay = None

		cursor = os.path.join(get_share_path(), "images", "menu-cursor.svg")
		self.cursor = Gtk.Image.new_from_file(cursor)
		self.cursor.set_name("osd-menu-cursor")

		self.parent = self.create_parent()
		self.f = Gtk.Fixed()
		self.f.add(self.parent)
		self.add(self.f)

		self._submenu = None
		self._scon = StickController()
		self._scon.connect("direction", self.on_stick_direction)
		self._is_submenu = False
		self._selected = None
		self._menuid = None
		self._use_cursor = False
		self._eh_ids = []
		self._control_with = STICK
		self._control_with_dpad = False
		self._confirm_with = "A"
		self._cancel_with = "B"

	def set_is_submenu(self):
		"""Marks menu as submenu. This changes behaviour of some methods,
		especially disables (un)locking of input stick and buttons.
		"""
		self._is_submenu = True

	def create_parent(self):
		v = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
		v.set_name("osd-menu")
		return v

	def pack_items(self, parent, items):
		for item in items:
			parent.pack_start(item.widget, True, True, 0)

	def use_daemon(self, d):
		"""Allows (re)using already existing DaemonManager instance in same process.
		use_config() should be be called before parse_argumets() if this is used.
		"""
		self.daemon = d
		if not self._is_submenu:
			self._connect_handlers()
			self.on_daemon_connected(self.daemon)

	def use_config(self, c):
		"""Allows reusing already existin Config instance in same process.
		Has to be called before parse_argumets()
		"""
		self.config = c

	def get_menuid(self):
		"""Returns ID of used menu.
		"""
		return self._menuid

	def get_selected_item_id(self):
		"""Returns ID of selected item or None if nothing is selected.
		"""
		if self._selected:
			return self._selected.id
		return None

	def _add_arguments(self):
		OSDWindow._add_arguments(self)
		self.argparser.add_argument(
			"--control-with",
			"-c",
			type=str,
			metavar="option",
			default=DEFAULT,
			choices=(DEFAULT, LEFT, RIGHT, STICK),
			help="which pad or stick should be used to navigate menu",
		)
		self.argparser.add_argument(
			"--confirm-with", type=str, metavar="button", default=DEFAULT, help="button used to confirm choice",
		)
		self.argparser.add_argument(
			"--cancel-with", type=str, metavar="button", default=DEFAULT, help="button used to cancel menu",
		)
		self.argparser.add_argument(
			"--confirm-with-release",
			action="store_true",
			help="confirm choice with button release instead of button press",
		)
		self.argparser.add_argument(
			"--cancel-with-release", action="store_true", help="cancel menu with button release instead of button press",
		)
		self.argparser.add_argument("--use-cursor", "-u", action="store_true", help="display and use cursor")
		self.argparser.add_argument("--size", type=int, help="sets prefered width or height")
		self.argparser.add_argument(
			"--feedback-amplitude",
			type=int,
			help="enables and sets power of feedback effect generated when active menu option is changed",
		)
		self.argparser.add_argument(
			"--from-profile", "-p", type=str, metavar="profile_file menu_name", help="load menu items from profile file",
		)
		self.argparser.add_argument(
			"--from-file", "-f", type=str, metavar="filename", help="load menu items from json file",
		)
		self.argparser.add_argument("--print-items", action="store_true", help="prints menu items to stdout")
		self.argparser.add_argument("items", type=str, nargs="*", metavar="id title", help="Menu items")

	@staticmethod
	def _get_on_screen_position(w):
		a = w.get_allocation()
		parent = w.get_parent()
		if parent:
			if isinstance(parent, Menu) and parent.get_window() is not None:
				x, y = parent.get_window().get_position()
			else:
				x, y = Menu._get_on_screen_position(parent)
			return a.x + x, a.y + y
		return a.x, a.y

	def parse_menu(self):
		if self.args.from_profile:
			try:
				self._menuid = self.args.items[0]
				self.items = MenuData.from_profile(self.args.from_profile, self._menuid)
			except OSError:
				print("%s: error: profile file not found" % (sys.argv[0]), file=sys.stderr)
				return False
			except ValueError:
				print("%s: error: menu not found" % (sys.argv[0]), file=sys.stderr)
				return False
		elif self.args.from_file:
			try:
				self._menuid = self.args.from_file
				self.items = MenuData.from_file(self.args.from_file)
			except:
				print("%s: error: failed to load menu file" % (sys.argv[0]), file=sys.stderr)
				return False
		else:
			try:
				self.items = MenuData.from_args(self.args.items)
				self._menuid = None
			except ValueError:
				print("%s: error: invalid number of arguments" % (sys.argv[0]), file=sys.stderr)
				return False
		return True

	def parse_argumets(self, argv):
		if not OSDWindow.parse_argumets(self, argv):
			return False
		if not self.parse_menu():
			return False
		if not self.config:
			self.config = Config()

		# Parse simpler arguments
		self._size = self.args.size

		# Create buttons that are displayed on screen
		items = self.items.generate(self)
		self.items = []
		for item in items:
			item.widget = self.generate_widget(item)
			if item.widget is not None:
				self.items.append(item)
		self.pack_items(self.parent, self.items)
		if len(self.items) == 0:
			print("%s: error: no items in menu" % (sys.argv[0]), file=sys.stderr)
			return False

		if self.args.print_items:
			max_id_len = max(*[len(x.id) for x in self.items])
			row_format = "{:>%s}:\t{}" % (max_id_len,)
			for item in self.items:
				print(row_format.format(item.id, item.label))
		return True

	def enable_cursor(self):
		if not self._use_cursor:
			self.f.add(self.cursor)
			self.f.show_all()
			self._use_cursor = True

	def generate_widget(self, item):
		"""Generates gtk widget for specified menutitem"""
		if isinstance(item, Separator) and item.label:
			widget = Gtk.Button.new_with_label(item.label)
			widget.set_relief(Gtk.ReliefStyle.NONE)
			widget.set_name("osd-menu-separator")
			return widget
		if isinstance(item, Separator):
			widget = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
			widget.set_name("osd-menu-separator")
			return widget
		widget = Gtk.Button.new_with_label(item.label)
		widget.set_relief(Gtk.ReliefStyle.NONE)
		if hasattr(widget.get_children()[0], "set_xalign"):
			widget.get_children()[0].set_xalign(0)
		else:
			widget.get_children()[0].set_halign(Gtk.Align.START)
		if isinstance(item, Submenu):
			item.callback = self.show_submenu
			label1 = widget.get_children()[0]
			label2 = Gtk.Label(_(">>"))
			label2.set_property("margin-left", 30)
			box = Gtk.Box(Gtk.Orientation.HORIZONTAL)
			widget.remove(label1)
			box.pack_start(label1, True, True, 1)
			box.pack_start(label2, False, True, 1)
			widget.add(box)
			widget.set_name("osd-menu-item")
		elif item.id is None:
			widget.set_name("osd-menu-dummy")
		else:
			widget.set_name("osd-menu-item")

		if isinstance(item.icon, Gio.FileIcon):
			icon_file = item.icon.get_file().get_path()
			has_colors = True
		elif isinstance(item.icon, Gio.ThemedIcon):
			icon = Gtk.IconTheme.get_default().choose_icon(item.icon.get_names(), 64, 0)
			icon_file = icon.get_filename() if icon else None
			has_colors = True
		else:
			icon_file, has_colors = find_icon(item.icon, self.PREFER_BW_ICONS)

		if icon_file:
			icon = MenuIcon(icon_file, has_colors)
			label = widget.get_children()[0]
			for c in [] + widget.get_children():
				widget.remove(c)
			box = Gtk.Box()
			box.pack_start(icon, False, True, 0)
			box.pack_start(label, True, True, 10)
			widget.add(box)

		return widget

	def select(self, index):
		if self._selected:
			self._selected.widget.set_name(self._selected.widget.get_name().replace("-selected", ""))
		if self.items[index].id:
			if self._selected != self.items[index]:
				if self.feedback and self.controller:
					self.controller.feedback(*self.feedback)
			self._selected = self.items[index]
			self._selected.widget.set_name(self._selected.widget.get_name() + "-selected")
			GLib.timeout_add(2, self._check_on_screen_position)
			return True
		return False

	def _check_on_screen_position(self, quick=False):
		if self.using_wlroots:
			return
		x, y = Menu._get_on_screen_position(self._selected.widget)
		try:
			m = self.get_window().get_display().get_monitor_at_window(self.get_window())
			assert m
			y_offset = m.get_geometry().y
			screen_height = m.get_geometry().height
		except:
			y_offset = 0
			screen_height = self.get_window().get_screen().get_height()
		y -= y_offset
		if y < 50:
			wx, wy = self.get_window().get_position()
			if quick:
				wy = 50 - (y - wy)
			else:
				wy += 5
				GLib.timeout_add(2, self._check_on_screen_position)
			self.get_window().move(wx, wy)
		if y > screen_height - 100:
			wx, wy = self.get_window().get_position()
			if quick:
				wy = screen_height - 100 - (y - wy)
			else:
				wy -= 5
				GLib.timeout_add(2, self._check_on_screen_position)
			self.get_window().move(wx, wy)

	def _connect_handlers(self):
		self._eh_ids += [
			(self.daemon, self.daemon.connect("dead", self.on_daemon_died)),
			(self.daemon, self.daemon.connect("error", self.on_daemon_died)),
			(self.daemon, self.daemon.connect("alive", self.on_daemon_connected)),
		]

	def run(self):
		self.daemon = DaemonManager()
		self._connect_handlers()
		OSDWindow.run(self)

	def show(self, *a):
		if not self.select(0):
			self.next_item(1)
		OSDWindow.show(self, *a)
		GLib.timeout_add(1, self._check_on_screen_position, True)

	def on_daemon_connected(self, *a):
		if not self.config:
			self.config = Config()
		self.controller = self.choose_controller(self.daemon)
		if self.controller is None or not self.controller.is_connected():
			# There is no controller connected to daemon
			self.on_failed_to_lock("Controller not connected")
			return
		self.use_controller(self.controller)

		self._eh_ids += [
			(self.controller, self.controller.connect("event", self.on_event)),
			(self.controller, self.controller.connect("lost", self.on_controller_lost)),
		]
		self.lock_inputs()

	def use_controller(self, controller):
		ccfg = self.config.get_controller_config(controller.get_id())
		self._control_with = getattr(self.args, "control_with", DEFAULT)
		self._cancel_with = getattr(self.args, "cancel_with", DEFAULT)
		if self._control_with == DEFAULT:
			self._control_with = ccfg["menu_control"]
		if self._cancel_with == DEFAULT:
			self._cancel_with = ccfg["menu_cancel"]

		self._confirm_with = getattr(self.args, "confirm_with", DEFAULT)
		if self._confirm_with == DEFAULT:
			self._confirm_with = ccfg["menu_confirm"]
		elif self._confirm_with == SAME:
			if self._control_with == RIGHT:
				self._confirm_with = SCButtons.RPADTOUCH.name
			else:
				self._confirm_with = SCButtons.LPADTOUCH.name

		if getattr(self.args, "use_cursor", False):
			# As special case, using LEFT pad on controller with
			# actual DPAD should not display cursor
			if self._control_with != LEFT or (controller.get_flags() & ControllerFlags.HAS_DPAD) == 0:
				self.enable_cursor()

		if getattr(self.args, "feedback_amplitude", None):
			side = "LEFT"
			if self._control_with == "RIGHT":
				side = "RIGHT"
			elif self._control_with == "STICK":
				side = "BOTH"
			self.feedback = side, int(self.args.feedback_amplitude)

	def lock_inputs(self):
		def success(*a):
			log.error("Sucessfully locked input")

		locks = [self._control_with, self._confirm_with, self._cancel_with]
		if self._control_with == "STICK":
			if self.controller.get_flags() & ControllerFlags.HAS_DPAD != 0:
				self._control_with_dpad = True
				locks += ["LEFT"]
		self.controller.lock(success, self.on_failed_to_lock, *locks)

	def quit(self, code=-2):
		if not self._is_submenu:
			if self.get_controller():
				self.get_controller().unlock_all()
			for source, eid in self._eh_ids:
				source.disconnect(eid)
			self._eh_ids = []
		OSDWindow.quit(self, code)

	def next_item(self, direction):
		"""Selects next menu item, based on self._direction"""
		start, i = -1, 0
		try:
			start = self.items.index(self._selected)
			i = start + direction
		except:
			pass
		while True:
			if i == start:
				# Cannot find valid menu item
				self.select(start)
				break
			if i >= len(self.items):
				i = 0
				GLib.timeout_add(1, self._check_on_screen_position, True)
				continue
			if i < 0:
				i = len(self.items) - 1
				GLib.timeout_add(1, self._check_on_screen_position, True)
				continue
			if self.select(i):
				# Not a separator
				break
			i += direction
			start = max(start, 0)

	def on_submenu_closed(self, *a):
		self.set_name("osd-menu")
		if self._submenu.get_exit_code() in (0, -2):
			self._menuid = self._submenu._menuid
			self._selected = self._submenu._selected
			self.quit(self._submenu.get_exit_code())
		self._submenu = None
		if self.using_wlroots:
			self.layer_shell.set_layer(self, self.layer_shell.Layer.OVERLAY)

	def show_submenu(self, trash, trash2, trash3, menuitem):
		"""Called when user chooses menu item pointing to submenu"""
		filename = find_menu(menuitem.filename)
		if filename:
			layer = None
			if self.using_wlroots:
				if self.layer_shell.is_supported():
					layer = self.layer_shell.Layer.OVERLAY
					self.layer_shell.set_layer(self, self.layer_shell.Layer.TOP)
			self._submenu = self.__class__(layer=layer)
			sub_pos = list(self.position)
			for i in (0, 1):
				sub_pos[i] = sub_pos[i] - self.SUBMENU_OFFSET if sub_pos[i] < 0 else sub_pos[i] + self.SUBMENU_OFFSET

			self._submenu.use_config(self.config)
			self._submenu.parse_argumets(
				[
					"menu.py",
					"-x",
					str(sub_pos[0]),
					"-y",
					str(sub_pos[1]),
					"--from-file",
					filename,
					"--control-with",
					self._control_with,
					"--confirm-with",
					self._confirm_with,
					"--cancel-with",
					self._cancel_with,
				],
			)
			self._submenu.set_is_submenu()
			self._submenu.use_daemon(self.daemon)
			self._submenu.use_controller(self.controller)
			self._submenu.controller = self.controller
			self._submenu.connect("destroy", self.on_submenu_closed)
			self._submenu.show()
			self.set_name("osd-menu-inactive")

	def _control_equals_cancel(self, daemon, x, y):
		"""Called by on_event in that very special case when both confirm_with
		and cancel_with are set to STICK.

		Separated because RadialMenu overrides on_event and still
		needs to call this.

		Returns True if menu was canceled.
		"""
		distance = sqrt(x * x + y * y)
		if distance < STICK_PAD_MAX / 8:
			self.quit(-1)
			return True
		return False

	def on_stick_direction(self, trash, x, y):
		if y != 0:
			self.next_item(y)

	def on_event(self, daemon, what, data):
		if self._submenu:
			return self._submenu.on_event(daemon, what, data)
		if what == self._control_with or (what == "LEFT" and self._control_with_dpad):
			x, y = data
			if self._use_cursor:
				# Special case, both confirm_with and cancel_with
				# can be set to STICK
				if self._cancel_with == STICK and self._control_with == STICK:
					if self._control_equals_cancel(daemon, x, y):
						return None

				pad_w = self.cursor.get_allocation().width * 0.5
				pad_h = self.cursor.get_allocation().height * 0.5
				max_w = self.get_allocation().width - 2 * pad_w
				max_h = self.get_allocation().height - 2 * pad_h

				x, y = circle_to_square(x / (STICK_PAD_MAX * 2.0), y / (STICK_PAD_MAX * 2.0))
				x = clamp(pad_w, (pad_w + max_w) * 0.5 + x * max_w, max_w - pad_w)
				y = clamp(pad_h, (pad_h + max_h) * 0.5 + y * max_h * -1, max_h - pad_h)
				self.f.move(self.cursor, int(x), int(y))

				# for i in self.items:
				# if point_in_gtkrect(i.widget.get_allocation(), x, y):
				# self.select(self.items.index(i))
			else:
				self._scon.set_stick(x, y)
		elif what == self._confirm_with:
			if data[0] == 0:  # Button released
				if self._selected and self._selected.callback:
					self._selected.callback(self, self.daemon, self.controller, self._selected)
				elif self._selected:
					self.quit(0)
				else:
					self.quit(-1)
		elif what == self._cancel_with:
			if data[0] == 0:  # Button released
				self.quit(-1)


class MenuIcon(Gtk.DrawingArea):
	"""Auti-sized, auto-recolored icon for menus"""

	def __init__(self, filename, has_colors=False):
		Gtk.DrawingArea.__init__(self)
		self.connect("size_allocate", self.on_size_allocate)
		self.has_colors = has_colors
		self.set_filename(filename)

	def set_filename(self, filename):
		if filename is None:
			self.pb = None
		else:
			self.pb = GdkPixbuf.Pixbuf.new_from_file(filename)

	def on_size_allocate(self, trash, allocation):
		if allocation.width < allocation.height:
			self.set_size_request(allocation.height, -1)

	def do_draw(self, cr):
		allocation = self.get_allocation()
		if allocation.width >= allocation.height:
			context = Gtk.Widget.get_style_context(self)
			Gtk.render_background(context, cr, 0, 0, allocation.width, allocation.height)
			if self.pb is None:
				# No icon set
				return
			scaled = self.pb.scale_simple(allocation.height, allocation.height, GdkPixbuf.InterpType.BILINEAR)
			surf = Gdk.cairo_surface_create_from_pixbuf(scaled, 1)
			if self.has_colors:
				cr.set_source_surface(surf, 1.0, 1.0)
				# allocation.height, allocation.height)
				cr.rectangle(0, 0, allocation.height, allocation.height)
			else:
				Gdk.cairo_set_source_rgba(cr, context.get_color(Gtk.StateFlags.NORMAL))
				cr.mask_surface(surf, 0, 0)
			cr.fill()
